# -*- coding: utf-8 -*-
"""MRI sampling functions.
"""
import numpy as np

from sigpy import util

__all__ = ["poisson", "radial", "spiral"]


def poisson(
    img_shape,
    accel,
    calib=(0, 0),
    dtype=complex,
    crop_corner=True,
    return_density=False,
    seed=0,
    max_attempts=30,
    tol=0.1,
):
    """Generate variable-density Poisson-disc sampling pattern.

    The function generates a variable density Poisson-disc sampling
    mask with density proportional to :math:`1 / (1 + s |r|)`,
    where :math:`r` represents the k-space radius, and :math:`s`
    represents the slope. A binary search is performed on the slope :math:`s`
    such that the resulting acceleration factor is close to the
    prescribed acceleration factor `accel`. The parameter `tol`
    determines how much they can deviate.

    Args:
        img_shape (tuple of ints): length-2 image shape.
        accel (float): Target acceleration factor. Must be greater than 1.
        calib (tuple of ints): length-2 calibration shape.
        dtype (Dtype): data type.
        crop_corner (bool): Toggle whether to crop sampling corners.
        seed (int): Random seed.
        max_attempts (float): maximum number of samples to reject in Poisson
           disc calculation.
        tol (float): Tolerance for how much the resulting acceleration can
            deviate form `accel`.

    Returns:
        array: Poisson-disc sampling mask.

    References:
        Bridson, Robert. "Fast Poisson disk sampling in arbitrary dimensions."
        SIGGRAPH sketches. 2007.

    """
    if accel <= 1:
        raise ValueError(f"accel must be greater than 1, got {accel}")

    if seed is not None:
        rand_state = np.random.get_state()

    ny, nx = img_shape
    y, x = np.mgrid[:ny, :nx]
    x = np.maximum(abs(x - img_shape[-1] / 2) - calib[-1] / 2, 0)
    x /= x.max()
    y = np.maximum(abs(y - img_shape[-2] / 2) - calib[-2] / 2, 0)
    y /= y.max()
    r = np.sqrt(x**2 + y**2)

    slope_max = max(nx, ny)
    slope_min = 0
    while slope_min < slope_max:
        slope = (slope_max + slope_min) / 2
        radius_x = np.clip((1 + r * slope) * nx / max(nx, ny), 1, None)
        radius_y = np.clip((1 + r * slope) * ny / max(nx, ny), 1, None)
        mask = _poisson(
            img_shape[-1],
            img_shape[-2],
            max_attempts,
            radius_x,
            radius_y,
            calib,
            seed,
        )
        if crop_corner:
            mask *= r < 1

        actual_accel = img_shape[-1] * img_shape[-2] / np.sum(mask)

        if abs(actual_accel - accel) < tol:
            break
        if actual_accel < accel:
            slope_min = slope
        else:
            slope_max = slope

    if abs(actual_accel - accel) >= tol:
        raise ValueError(f"Cannot generate mask to satisfy accel={accel}.")

    mask = mask.reshape(img_shape).astype(dtype)

    if seed is not None:
        np.random.set_state(rand_state)

    return mask


def radial(coord_shape, img_shape, golden=True, dtype=float):
    """Generate radial trajectory.

    Args:
        coord_shape (tuple of ints): coordinates of shape [ntr, nro, ndim],
            where ntr is the number of TRs, nro is the number of readout,
            and ndim is the number of dimensions.
        img_shape (tuple of ints): image shape.
        golden (bool): golden angle ordering.
        dtype (Dtype): data type.

    Returns:
        array: radial coordinates.

    References:
        1. An Optimal Radial Profile Order Based on the Golden
        Ratio for Time-Resolved MRI
        Stefanie Winkelmann, Tobias Schaeffter, Thomas Koehler,
        Holger Eggers, and Olaf Doessel. TMI 2007.
        2. Temporal stability of adaptive 3D radial MRI using
        multidimensional golden means
        Rachel W. Chan, Elizabeth A. Ramsay, Charles H. Cunningham,
        and Donald B. Plewes. MRM 2009.

    """
    if len(img_shape) != coord_shape[-1]:
        raise ValueError(
            "coord_shape[-1] must match len(img_shape), "
            "got {} and {}".format(coord_shape[-1], len(img_shape))
        )

    ntr, nro, ndim = coord_shape
    if ndim == 2:
        if golden:
            phi = np.pi * (3 - 5**0.5)
        else:
            phi = 2 * np.pi / ntr

        n, r = np.mgrid[:ntr, : 0.5 : 0.5 / nro]

        theta = n * phi
        coord = np.zeros((ntr, nro, 2))
        coord[:, :, -1] = r * np.cos(theta)
        coord[:, :, -2] = r * np.sin(theta)

    elif ndim == 3:
        if golden:
            phi1 = 0.465571231876768
            phi2 = 0.682327803828019
        else:
            raise NotImplementedError

        n, r = np.mgrid[:ntr, : 0.5 : 0.5 / nro]

        beta = np.arccos(2 * ((n * phi1) % 1) - 1)
        alpha = 2 * np.pi * ((n * phi2) % 1)

        coord = np.zeros((ntr, nro, 3))
        coord[:, :, -1] = r * np.sin(beta) * np.cos(alpha)
        coord[:, :, -2] = r * np.sin(beta) * np.sin(alpha)
        coord[:, :, -3] = r * np.cos(beta)
    else:
        raise ValueError("coord_shape[-1] must be 2 or 3, got {}".format(ndim))

    return (coord * img_shape[-ndim:]).astype(dtype)


@util.fallback_jit(nopython=True, cache=True)  # pragma: no cover
def _poisson(nx, ny, max_attempts, radius_x, radius_y, calib, seed=None):
    mask = np.zeros((ny, nx))

    # Add calibration region
    mask[
        int(ny / 2 - calib[-2] / 2) : int(ny / 2 + calib[-2] / 2),
        int(nx / 2 - calib[-1] / 2) : int(nx / 2 + calib[-1] / 2),
    ] = 1

    if seed is not None:
        np.random.seed(int(seed))

    # initialize active list
    pxs = np.empty(nx * ny, np.int32)
    pys = np.empty(nx * ny, np.int32)
    pxs[0] = np.random.randint(0, nx)
    pys[0] = np.random.randint(0, ny)
    num_actives = 1
    while num_actives > 0:
        i = np.random.randint(0, num_actives)
        px = pxs[i]
        py = pys[i]
        rx = radius_x[py, px]
        ry = radius_y[py, px]

        # Attempt to generate point
        done = False
        k = 0
        while not done and k < max_attempts:
            # Generate point randomly from r and 2 * r
            v = (np.random.random() * 3 + 1) ** 0.5
            t = 2 * np.pi * np.random.random()
            qx = px + v * rx * np.cos(t)
            qy = py + v * ry * np.sin(t)

            # Reject if outside grid or close to other points
            if qx >= 0 and qx < nx and qy >= 0 and qy < ny:
                startx = max(int(qx - rx), 0)
                endx = min(int(qx + rx + 1), nx)
                starty = max(int(qy - ry), 0)
                endy = min(int(qy + ry + 1), ny)

                done = True
                for x in range(startx, endx):
                    for y in range(starty, endy):
                        if mask[y, x] == 1 and (
                            ((qx - x) / radius_x[y, x]) ** 2
                            + ((qy - y) / (radius_y[y, x])) ** 2
                            < 1
                        ):
                            done = False
                            break

            k += 1

        # Add point if done else remove from active list
        if done:
            pxs[num_actives] = qx
            pys[num_actives] = qy
            mask[int(qy), int(qx)] = 1
            num_actives += 1
        else:
            pxs[i] = pxs[num_actives - 1]
            pys[i] = pys[num_actives - 1]
            num_actives -= 1

    return mask


def spiral(fov, N, f_sampling, R, ninterleaves, alpha, gm, sm, gamma=2.678e8):
    """Generate variable density spiral trajectory.

    Args:
        fov (float): field of view in meters.
        N (int): effective matrix shape.
        f_sampling (float): undersampling factor in freq encoding direction.
        R (float): undersampling factor.
        ninterleaves (int): number of spiral interleaves
        alpha (float): variable density factor
        gm (float): maximum gradient amplitude (T/m)
        sm (float): maximum slew rate (T/m/s)
        gamma (float): gyromagnetic ratio in rad/T/s

    Returns:
        array: spiral coordinates.

    References:
        Dong-hyun Kim, Elfar Adalsteinsson, and Daniel M. Spielman.
        'Simple Analytic Variable Density Spiral Design.' MRM 2003.

    """
    res = fov / N

    lam = 0.5 / res  # in m**(-1)
    n = 1 / (1 - (1 - ninterleaves * R / fov / lam) ** (1 / alpha))
    w = 2 * np.pi * n
    Tea = lam * w / gamma / gm / (alpha + 1)  # in s
    Tes = np.sqrt(lam * w**2 / sm / gamma) / (alpha / 2 + 1)  # in s
    Ts2a = (
        Tes ** ((alpha + 1) / (alpha / 2 + 1))
        * (alpha / 2 + 1)
        / Tea
        / (alpha + 1)
    ) ** (
        1 + 2 / alpha
    )  # in s

    if Ts2a < Tes:
        tautrans = (Ts2a / Tes) ** (1 / (alpha / 2 + 1))

        def tau(t):
            return (t / Tes) ** (1 / (alpha / 2 + 1)) * (0 <= t) * (
                t <= Ts2a
            ) + ((t - Ts2a) / Tea + tautrans ** (alpha + 1)) ** (
                1 / (alpha + 1)
            ) * (
                t > Ts2a
            ) * (
                t <= Tea
            ) * (
                Tes >= Ts2a
            )

        Tend = Tea
    else:

        def tau(t):
            return (t / Tes) ** (1 / (alpha / 2 + 1)) * (0 <= t) * (t <= Tes)

        Tend = Tes

    def k(t):
        return lam * tau(t) ** alpha * np.exp(w * tau(t) * 1j)

    dt = Tea * 1e-4  # in s

    Dt = dt * f_sampling / fov / abs(k(Tea) - k(Tea - dt))  # in s

    t = np.linspace(0, Tend, int(Tend / Dt))
    kt = k(t)  # in rad

    # generating cloned interleaves
    k = kt
    for i in range(1, ninterleaves):
        k = np.hstack((k, kt[0:] * np.exp(2 * np.pi * 1j * i / ninterleaves)))

    k = np.stack((np.real(k), np.imag(k)), axis=1)

    return k



# -*- coding: utf-8 -*-
"""Thresholding functions.
"""
import numpy as np

from sigpy import backend, config, util

__all__ = [
    "soft_thresh",
    "hard_thresh",
    "l1_proj",
    "l2_proj",
    "linf_proj",
    "psd_proj",
]


def soft_thresh(lamda, input):
    r"""Soft threshold.

    Performs:

    .. math::
        (| x | - \lambda)_+  \text{sgn}(x)

    Args:
        lamda (float, or array): Threshold parameter.
        input (array)

    Returns:
        array: soft-thresholded result.

    """
    device = backend.get_device(input)
    xp = device.xp
    if xp == np:
        return _soft_thresh(lamda, input)
    else:  # pragma: no cover
        if np.isscalar(lamda):
            lamda = backend.to_device(lamda, device)

        return _soft_thresh_cuda(lamda, input)


def hard_thresh(lamda, input):
    """Hard threshold.

    Args:
        lamda (float, or array): Threshold parameter.
        input (array)

    Returns:
        array: hard-thresholded result.

    """
    device = backend.get_device(input)
    xp = device.xp
    if xp == np:
        return _hard_thresh(lamda, input)
    else:  # pragma: no cover
        if np.isscalar(lamda):
            lamda = backend.to_device(lamda, device)

        return _hard_thresh_cuda(lamda, input)


def l1_proj(eps, input):
    """Projection onto L1 ball.

    Args:
        eps (float, or array): L1 ball scaling.
        input (array)

    Returns:
        array: Result.

    References:
        J. Duchi, S. Shalev-Shwartz, and Y. Singer, "Efficient projections onto
        the l1-ball for learning in high dimensions" 2008.

    """
    xp = backend.get_array_module(input)
    shape = input.shape
    input = input.ravel()

    if xp.linalg.norm(input, 1) < eps:
        return input
    else:
        size = len(input)
        s = xp.sort(xp.abs(input))[::-1]
        st = (xp.cumsum(s) - eps) / (xp.arange(size) + 1)
        idx = xp.flatnonzero((s - st) > 0).max()
        return soft_thresh(st[idx], input.reshape(shape))


def l2_proj(eps, input, axes=None):
    """Projection onto L2 ball.

    Args:
        eps (float, or array): L2 ball scaling.
        input (array)

    Returns:
        array: Result.

    """
    axes = util._normalize_axes(axes, input.ndim)

    xp = backend.get_array_module(input)
    norm = xp.sum(xp.abs(input) ** 2, axis=axes, keepdims=True) ** 0.5
    mask = norm < eps
    output = mask * input + (1 - mask) * (eps * input / (norm + mask))

    return output


def linf_proj(eps, input, bias=None):
    """Projection onto L-infinity ball.

    Args:
        eps (float, or array): l-infinity ball scaling.
        input (array)

    Returns:
        array: Result.

    """
    if bias is not None:
        input = input - bias

    output = input - soft_thresh(eps, input)

    if bias is not None:
        output += bias

    return output


def psd_proj(input):
    """Projection onto postiive semi-definite matrices.

    Args:
        input (array): a two-dimensional matrix.

    Returns:
        array: Result.

    """
    xp = backend.get_array_module(input)
    w, v = xp.linalg.eig((input + xp.conj(input).T) / 2)
    w[w < 0] = 0
    return (v * w) @ v.conjugate().T


@util.fallback_vectorize  # pragma: no cover
def _soft_thresh(lamda, input):
    abs_input = abs(input)
    if abs_input == 0:
        sign = 0
    else:
        sign = input / abs_input

    mag = abs_input - lamda
    mag = (abs(mag) + mag) / 2

    return mag * sign


@util.fallback_vectorize  # pragma: no cover
def _hard_thresh(lamda, input):
    abs_input = abs(input)
    if abs_input > lamda:
        return input
    else:
        return 0


if config.cupy_enabled:  # pragma: no cover
    import cupy as cp

    _soft_thresh_cuda = cp.ElementwiseKernel(
        "S lamda, T input",
        "T output",
        """
        S abs_input = abs(input);
        T sign;
        if (abs_input == 0)
            sign = 0;
        else
            sign = input / (T) abs_input;
        S mag = abs_input - lamda;
        mag = (abs(mag) + mag) / 2.;

        output = (T) mag * sign;
        """,
        name="soft_thresh",
    )

    _hard_thresh_cuda = cp.ElementwiseKernel(
        "S lamda, T input",
        "T output",
        """
        S abs_input = abs(input);
        if (abs_input > lamda)
            output = input;
        else
            output = 0;
        """,
        name="hard_thresh",
    )



# -*- coding: utf-8 -*-
"""Utility functions.
"""
import numpy as np

from sigpy import backend, config


__all__ = [
    "prod",
    "vec",
    "split",
    "rss",
    "resize",
    "flip",
    "circshift",
    "downsample",
    "upsample",
    "dirac",
    "randn",
    "triang",
    "hanning",
    "monte_carlo_sure",
    "axpy",
    "xpay",
    "leja",
]


def _normalize_axes(axes, ndim):
    if axes is None:
        return tuple(range(ndim))
    else:
        return tuple(a % ndim for a in sorted(axes))


def _normalize_shape(shape):
    if isinstance(shape, int):
        return (shape,)
    else:
        return tuple(shape)


def _expand_shapes(*shapes):
    shapes = [list(shape) for shape in shapes]
    max_ndim = max(len(shape) for shape in shapes)
    shapes_exp = [[1] * (max_ndim - len(shape)) + shape for shape in shapes]

    return tuple(shapes_exp)


def _check_same_dtype(*arrays):
    dtype = arrays[0].dtype
    for a in arrays:
        if a.dtype != dtype:
            raise TypeError(
                "inputs dtype mismatch, got {a_dtype}, and {dtype}.".format(
                    a_dtype=a.dtype, dtype=dtype
                )
            )


def prod(shape):
    """Computes product of shape.

    Args:
        shape (tuple or list): shape.

    Returns:
        Product.

    """
    return np.prod(shape, dtype=np.int64)


def vec(inputs):
    """Vectorize inputs.

    Args:
        shape (tuple or list): shape.

    Returns:
        array: Vectorized result.
    """
    xp = backend.get_array_module(inputs[0])
    return xp.concatenate([i.ravel() for i in inputs])


def split(vec, oshapes):
    """Split input into specified output shapes.

    Args:
        oshapes (list of tuple of ints): Output shapes.

    Returns:
        list of arrays: Splitted outputs.
    """
    outputs = []
    for oshape in oshapes:
        osize = prod(oshape)
        outputs.append(vec[:osize].reshape(oshape))
        vec = vec[osize:]

    return outputs


def rss(input, axes=(0,)):
    """Root sum of squares.

    Args:
        input (array): Input array.
        axes (None or tuple of ints): Axes to perform operation.

    Returns:
        array: Result.
    """
    xp = backend.get_array_module(input)
    return xp.sum(xp.abs(input) ** 2, axis=axes) ** 0.5


def resize(input, oshape, ishift=None, oshift=None):
    """Resize with zero-padding or cropping.

    Args:
        input (array): Input array.
        oshape (tuple of ints): Output shape.
        ishift (None or tuple of ints): Input shift.
        oshift (None or tuple of ints): Output shift.

    Returns:
        array: Zero-padded or cropped result.
    """

    ishape1, oshape1 = _expand_shapes(input.shape, oshape)

    if ishape1 == oshape1:
        return input.reshape(oshape)

    if ishift is None:
        ishift = [max(i // 2 - o // 2, 0) for i, o in zip(ishape1, oshape1)]

    if oshift is None:
        oshift = [max(o // 2 - i // 2, 0) for i, o in zip(ishape1, oshape1)]

    copy_shape = [
        min(i - si, o - so)
        for i, si, o, so in zip(ishape1, ishift, oshape1, oshift)
    ]
    islice = tuple([slice(si, si + c) for si, c in zip(ishift, copy_shape)])
    oslice = tuple([slice(so, so + c) for so, c in zip(oshift, copy_shape)])

    xp = backend.get_array_module(input)
    output = xp.zeros(oshape1, dtype=input.dtype)
    input = input.reshape(ishape1)
    output[oslice] = input[islice]

    return output.reshape(oshape)


def flip(input, axes=None):
    """Flip input.

    Args:
        input (array): Input array.
        axes (None or tuple of ints): Axes to perform operation.

    Returns:
        array: Flipped result.
    """

    axes = _normalize_axes(axes, input.ndim)

    slc = []
    for d in range(input.ndim):
        if d in axes:
            slc.append(slice(None, None, -1))
        else:
            slc.append(slice(None))

    slc = tuple(slc)
    output = input[slc]

    return output


def circshift(input, shifts, axes=None):
    """Circular shift input.

    Args:
        input (array): Input array.
        shifts (tuple of ints): Shifts.
        axes (None or tuple of ints): Axes to perform operation.

    Returns:
        array: Result.
    """

    if axes is None:
        axes = range(input.ndim)

    assert len(axes) == len(shifts)
    xp = backend.get_array_module(input)

    for axis, shift in zip(axes, shifts):
        input = xp.roll(input, shift, axis=axis)

    return input


def downsample(input, factors, shift=None):
    """Downsample input.

    Args:
        input (array): Input array.
        factors (tuple of ints): Downsampling factors.
        shifts (None or tuple of ints): Shifts.

    Returns:
        array: Result.
    """

    if shift is None:
        shift = [0] * len(factors)

    slc = tuple(slice(s, None, f) for s, f in zip(shift, factors))
    return input[slc]


def upsample(input, oshape, factors, shift=None):
    """Upsample input.

    Args:
        input (array): Input array.
        factors (tuple of ints): Upsampling factors.
        shifts (None or tuple of ints): Shifts.

    Returns:
        array: Result.
    """

    if shift is None:
        shift = [0] * len(factors)

    slc = tuple(slice(s, None, f) for s, f in zip(shift, factors))

    xp = backend.get_array_module(input)
    output = xp.zeros(oshape, dtype=input.dtype)
    output[slc] = input

    return output


def dirac(shape, dtype=float, device=backend.cpu_device):
    """Create Dirac delta.

    Args:
        shape (tuple of ints): Output shape.
        dtype (Dtype): Output data-type.
        device (Device): Output device.

    Returns:
        array: Dirac delta array.
    """

    device = backend.Device(device)
    xp = device.xp

    with device:
        return resize(xp.ones([1], dtype=dtype), shape)


def randn(shape, scale=1, dtype=float, device=backend.cpu_device):
    """Create random Gaussian array.

    Args:
        shape (tuple of ints): Output shape.
        scale (float): Standard deviation.
        dtype (Dtype): Output data-type.
        device (Device): Output device.

    Returns:
        array: Random Gaussian array.

    """
    device = backend.Device(device)
    xp = device.xp

    with device:
        if np.issubdtype(dtype, np.complexfloating):
            real_dtype = np.array([], dtype=dtype).real.dtype
            real_shape = tuple(shape) + (2,)
            output = xp.random.normal(size=real_shape, scale=scale / 2**0.5)
            output = output.astype(real_dtype)
            output = output.view(dtype=dtype).reshape(shape)
            return output
        else:
            return xp.random.normal(size=shape, scale=scale).astype(dtype)


def triang(shape, dtype=float, device=backend.cpu_device):
    """Create multi-dimensional triangular window.

    Args:
        shape (tuple of ints): Output shape.
        dtype (Dtype): Output data-type.
        device (Device): Output device.

    Returns:
        array: triangular filter.

    """
    device = backend.Device(device)
    xp = device.xp
    shape = _normalize_shape(shape)
    with device:
        window = xp.ones(shape, dtype=dtype)
        for n, i in enumerate(shape[::-1]):
            x = xp.arange(i, dtype=dtype)
            w = 1 - xp.abs(x - i // 2 + ((i + 1) % 2) / 2) / ((i + 1) // 2)
            window *= w.reshape([i] + [1] * n)

    return window


def hanning(shape, dtype=float, device=backend.cpu_device):
    """Create multi-dimensional hanning window.

    Args:
        shape (tuple of ints): Output shape.
        dtype (Dtype): Output data-type.
        device (Device): Output device.

    Returns:
        array: hanning filter.

    """
    device = backend.Device(device)
    xp = device.xp
    shape = _normalize_shape(shape)
    with device:
        window = xp.ones(shape, dtype=dtype)
        for n, i in enumerate(shape[::-1]):
            x = xp.arange(i, dtype=dtype)
            w = 0.5 - 0.5 * xp.cos(2 * np.pi * x / max(1, (i - (i % 2))))
            window *= w.reshape([i] + [1] * n)

    return window


def monte_carlo_sure(f, y, sigma, eps=1e-10):
    """Monte Carlo Stein Unbiased Risk Estimator (SURE).

    Monte carlo SURE assumes the observation y = x + e,
    where e is a white Gaussian array with standard deviation sigma.
    Monte carlo SURE provides an unbiased estimate of mean-squared error, ie:
    1 / n || f(y) - x ||_2^2

    Args:
        f (function): x -> f(x).
        y (array): observed measurement.
        sigma (float): noise standard deviation.

    Returns:
       float: SURE.

    References:
        Ramani, S., Blu, T. and Unser, M. 2008.
        Monte-Carlo Sure: A Black-Box Optimization of Regularization Parameters
        for General Denoising Algorithms. IEEE Transactions on Image Processing
        17, 9 (2008), 1540-1554.
    """
    device = backend.get_device(y)
    xp = device.xp

    n = y.size
    f_y = f(y)
    b = randn(y.shape, dtype=y.dtype, device=device)
    divf_y = xp.real(xp.vdot(b, (f(y + eps * b) - f_y))) / eps
    sure = (
        xp.mean(xp.abs(y - f_y) ** 2)
        - sigma**2
        + 2 * sigma**2 * divf_y / n
    )

    return sure


def leja(x):
    """Perform leja ordering of roots of a polynomial.

    Orders roots in a way suitable to accurately compute polynomial
    coefficients.

    Args:
        x (array): roots to be ordered.

    Returns:
        array: ordered roots.

    References:
        Lang, M. and B. Frenzel. 1993.
        A New and Efficient Program for Finding All Polynomial Roots. Rice
        University ECE Technical Report, no. TR93-08, 1993.
    """

    n = np.size(x)
    # duplicate roots to n+1 rows
    a = np.tile(np.reshape(x, (1, n)), (n + 1, 1))
    # take abs of first row
    a[0, :] = np.abs(a[0, :])

    tmp = np.zeros(n + 1, dtype=complex)

    # find index of max abs value
    ind = np.argmax(a[0, :])
    if ind != 0:
        tmp[:] = a[:, 0]
        a[:, 0] = a[:, ind]
        a[:, ind] = tmp

    x_out = np.zeros(n, dtype=complex)
    x_out[0] = a[n - 1, 0]  # first entry of last row
    a[1, 1:] = np.abs(a[1, 1:] - x_out[0])

    foo = a[0, 0:n]

    for ll in range(1, n - 1):
        foo = np.multiply(foo, a[ll, :])
        ind = np.argmax(foo[ll:])
        ind = ind + ll
        if ll != ind:
            tmp[:] = a[:, ll]
            a[:, ll] = a[:, ind]
            a[:, ind] = tmp
            # also swap inds in foo
            tmp[0] = foo[ll]
            foo[ll] = foo[ind]
            foo[ind] = tmp[0]
        x_out[ll] = a[n - 1, ll]
        a[ll + 1, (ll + 1) : n] = np.abs(a[ll + 1, (ll + 1) :] - x_out[ll])

    x_out = a[n, :]

    return x_out


def axpy(y, a, x):
    """Compute y = a * x + y.

    Args:
        y (array): Output array.
        a (scalar or array): Input scalar.
        x (array): Input array.

    """
    y += a * x


def xpay(y, a, x):
    """Compute y = x + a * y.

    Args:
        y (array): Output array.
        a (scalar or array): Input scalar.
        x (array): Input array.
    """
    y *= a
    y += x


def fallback_jit(*jit_args, **jit_kwargs):
    def decorator(func):
        if config.numba_enabled:
            import numba as nb # noqa
            return nb.jit(*jit_args, **jit_kwargs)(func)
        return func
    return decorator


def fallback_vectorize(*jit_args, **jit_kwargs):
    def decorator(func):
        if config.numba_enabled:
            import numba as nb # noqa
            return nb.vectorize(*jit_args, **jit_kwargs)(func)
        return func
    return decorator



# -*- coding: utf-8 -*-
"""Convolution functions with  multi-dimension, and multi-channel support.

"""
import numpy as np
import scipy.signal as signal

from sigpy import backend, config, util

__all__ = ["convolve", "convolve_data_adjoint", "convolve_filter_adjoint"]


def convolve(data, filt, mode="full", strides=None, multi_channel=False):
    r"""Convolution that supports multi-dimensional and multi-channel inputs.

    This function follows the signal processing definition of convolution.
    Note that the cuDNN version only supports inputs with D=1, 2 or 3.

    Args:
        data (array): data array of shape:
            :math:`[..., m_1, ..., m_D]` if multi_channel is False,
            :math:`[..., c_i, m_1, ..., m_D]` otherwise.
        filt (array): filter array of shape:
            :math:`[n_1, ..., n_D]` if multi_channel is False
            :math:`[c_o, c_i, n_1, ..., n_D]` otherwise.
        mode (str): {'full', 'valid'}.
        strides (None or tuple of ints): convolution strides of length D.
        multi_channel (bool): specify if input/output has multiple channels.

    Returns:
        array: output array of shape:
            :math:`[..., p_1, ..., p_D]` if multi_channel is False,
            :math:`[..., c_o, p_1, ..., p_D]` otherwise.

    """
    xp = backend.get_array_module(data)
    if xp == np:
        output = _convolve(
            data, filt, mode=mode, strides=strides, multi_channel=multi_channel
        )
    else:  # pragma: no cover
        if config.cudnn_enabled:
            if np.issubdtype(data.dtype, np.floating):
                output = _convolve_cuda(
                    data,
                    filt,
                    mode=mode,
                    strides=strides,
                    multi_channel=multi_channel,
                )
            else:
                output = _complex(
                    _convolve_cuda,
                    data,
                    filt,
                    mode=mode,
                    strides=strides,
                    multi_channel=multi_channel,
                )
        else:
            raise RuntimeError(
                "cudnn must be installed to perform convolution on GPU."
            )

    return output


def convolve_data_adjoint(
    output, filt, data_shape, mode="full", strides=None, multi_channel=False
):
    """Adjoint convolution operation with respect to data.

    Note that the cuDNN version only supports inputs with D=1, 2 or 3.

    Args:
        output (array): output array of shape
            :math:`[..., p_1, ..., p_D]` if multi_channel is False,
            :math:`[..., c_o, p_1, ..., p_D]` otherwise.
        filt (array): filter array of shape
            :math:`[n_1, ..., n_D]` if multi_channel is False
            :math:`[c_o, c_i, n_1, ..., n_D]` otherwise.
        mode (str): {'full', 'valid'}.
        strides (None or tuple of ints): convolution strides of length D.
        multi_channel (bool): specify if input/output has multiple channels.
        multi_channel (bool): specify if data/output has multiple channels.
        mode (str): {'full', 'valid'}.

    Returns:
        array: data array of shape
            :math:`[..., m_1, ..., m_D]` if multi_channel is False,
            :math:`[..., c_i, m_1, ..., m_D]` otherwise.

    """
    data_shape = tuple(data_shape)

    xp = backend.get_array_module(output)
    if xp == np:
        data = _convolve_data_adjoint(
            output,
            filt,
            data_shape,
            mode=mode,
            strides=strides,
            multi_channel=multi_channel,
        )
    else:  # pragma: no cover
        if config.cudnn_enabled:
            if np.issubdtype(output.dtype, np.floating):
                data = _convolve_data_adjoint_cuda(
                    output,
                    filt,
                    data_shape,
                    mode=mode,
                    strides=strides,
                    multi_channel=multi_channel,
                )
            else:
                data = _complex(
                    _convolve_data_adjoint_cuda,
                    output,
                    filt.conj(),
                    data_shape,
                    mode=mode,
                    strides=strides,
                    multi_channel=multi_channel,
                )
        else:
            raise RuntimeError(
                "cudnn must be installed to perform convolution on GPU."
            )

    return data


def convolve_filter_adjoint(
    output, data, filt_shape, mode="full", strides=None, multi_channel=False
):
    """Adjoint convolution operation with respect to filter.

    Note that the cuDNN version only supports inputs with D=1, 2 or 3.

    Args:
        output (array): output array of shape:
            :math:`[..., p_1, ..., p_D]` if multi_channel is False,
            :math:`[..., c_o, p_1, ..., p_D]` otherwise.
        data (array): data array of shape:
            :math:`[..., m_1, ..., m_D]` if multi_channel is False,
            :math:`[..., c_i, m_1, ..., m_D]` otherwise.
        mode (str): {'full', 'valid'}.
        strides (None or tuple of ints): convolution strides of length D.
        multi_channel (bool): specify if input/output has multiple channels.

    Returns:
        array: filter array of shape:
            :math:`[n_1, ..., n_D]` if multi_channel is False
            :math:`[c_o, c_i, n_1, ..., n_D]` otherwise.

    """
    filt_shape = tuple(filt_shape)
    xp = backend.get_array_module(data)
    if xp == np:
        filt = _convolve_filter_adjoint(
            output,
            data,
            filt_shape,
            mode=mode,
            strides=strides,
            multi_channel=multi_channel,
        )
    else:  # pragma: no cover
        if config.cudnn_enabled:
            if np.issubdtype(output.dtype, np.floating):
                filt = _convolve_filter_adjoint_cuda(
                    output,
                    data,
                    filt_shape,
                    mode=mode,
                    strides=strides,
                    multi_channel=multi_channel,
                )
            else:
                filt = _complex(
                    _convolve_filter_adjoint_cuda,
                    output,
                    data.conj(),
                    filt_shape,
                    mode=mode,
                    strides=strides,
                    multi_channel=multi_channel,
                )
        else:
            raise RuntimeError(
                "cudnn must be installed to perform convolution on GPU."
            )

    return filt


def _get_convolve_params(data_shape, filt_shape, mode, strides, multi_channel):
    D = len(filt_shape) - 2 * multi_channel
    m = tuple(data_shape[-D:])
    n = tuple(filt_shape[-D:])
    b = tuple(data_shape[: -D - multi_channel])
    B = util.prod(b)

    if multi_channel:
        if filt_shape[-D - 1] != data_shape[-D - 1]:
            raise ValueError(
                "Data channel mismatch, "
                "got {} from data and {} from filt.".format(
                    data_shape[-D - 1], filt_shape[-D - 1]
                )
            )

        c_i = filt_shape[-D - 1]
        c_o = filt_shape[-D - 2]
    else:
        c_i = 1
        c_o = 1

    if strides is None:
        s = (1,) * D
    else:
        if len(strides) != D:
            raise ValueError("Strides must have length {}.".format(D))

        s = tuple(strides)

    if mode == "full":
        p = tuple(
            (m_d + n_d - 1 + s_d - 1) // s_d for m_d, n_d, s_d in zip(m, n, s)
        )
    elif mode == "valid":
        if any(m_d >= n_d for m_d, n_d in zip(m, n)) and any(
            m_d < n_d for m_d, n_d in zip(m, n)
        ):
            raise ValueError(
                "In valid mode, either data or filter must be "
                "at least as large as the other in every axis."
            )

        p = tuple(
            (m_d - n_d + 1 + s_d - 1) // s_d for m_d, n_d, s_d in zip(m, n, s)
        )
    else:
        raise ValueError("Invalid mode, got {}".format(mode))

    return D, b, B, m, n, s, c_i, c_o, p


def _convolve(data, filt, mode="full", strides=None, multi_channel=False):
    D, b, B, m, n, s, c_i, c_o, p = _get_convolve_params(
        data.shape, filt.shape, mode, strides, multi_channel
    )

    # Normalize shapes.
    data = data.reshape((B, c_i) + m)
    filt = filt.reshape((c_o, c_i) + n)
    output = np.zeros((B, c_o) + p, dtype=data.dtype)
    slc = tuple(slice(None, None, s_d) for s_d in s)

    for k in range(B):
        for j in range(c_o):
            for i in range(c_i):
                output[k, j] += signal.convolve(
                    data[k, i], filt[j, i], mode=mode
                )[slc]

    # Reshape.
    if multi_channel:
        output = output.reshape(b + (c_o,) + p)
    else:
        output = output.reshape(b + p)

    return output


def _convolve_data_adjoint(
    output, filt, data_shape, mode="full", strides=None, multi_channel=False
):
    D, b, B, m, n, s, c_i, c_o, p = _get_convolve_params(
        data_shape, filt.shape, mode, strides, multi_channel
    )

    # Normalize shapes.
    output = output.reshape((B, c_o) + p)
    filt = filt.reshape((c_o, c_i) + n)
    data = np.zeros((B, c_i) + m, dtype=output.dtype)
    slc = tuple(slice(None, None, s_d) for s_d in s)
    if mode == "full":
        output_kj = np.zeros(
            [m_d + n_d - 1 for m_d, n_d in zip(m, n)], dtype=output.dtype
        )
        adjoint_mode = "valid"
    elif mode == "valid":
        output_kj = np.zeros(
            [max(m_d, n_d) - min(m_d, n_d) + 1 for m_d, n_d in zip(m, n)],
            dtype=output.dtype,
        )
        if all(m_d >= n_d for m_d, n_d in zip(m, n)):
            adjoint_mode = "full"
        else:
            adjoint_mode = "valid"

    for k in range(B):
        for j in range(c_o):
            for i in range(c_i):
                output_kj[slc] = output[k, j]
                data[k, i] += signal.correlate(
                    output_kj, filt[j, i], mode=adjoint_mode
                )

    # Reshape.
    data = data.reshape(data_shape)
    return data


def _convolve_filter_adjoint(
    output, data, filt_shape, mode="full", strides=None, multi_channel=False
):
    D, b, B, m, n, s, c_i, c_o, p = _get_convolve_params(
        data.shape, filt_shape, mode, strides, multi_channel
    )

    # Normalize shapes.
    data = data.reshape((B, c_i) + m)
    output = output.reshape((B, c_o) + p)
    slc = tuple(slice(None, None, s_d) for s_d in s)
    if mode == "full":
        output_kj = np.zeros(
            [m_d + n_d - 1 for m_d, n_d in zip(m, n)], dtype=output.dtype
        )
        adjoint_mode = "valid"
    elif mode == "valid":
        output_kj = np.zeros(
            [max(m_d, n_d) - min(m_d, n_d) + 1 for m_d, n_d in zip(m, n)],
            dtype=output.dtype,
        )
        if all(m_d >= n_d for m_d, n_d in zip(m, n)):
            adjoint_mode = "valid"
        else:
            adjoint_mode = "full"

    filt = np.zeros((c_o, c_i) + n, dtype=output.dtype)
    for k in range(B):
        for j in range(c_o):
            for i in range(c_i):
                output_kj[slc] = output[k, j]
                filt[j, i] += signal.correlate(
                    output_kj, data[k, i], mode=adjoint_mode
                )

    # Reshape.
    filt = filt.reshape(filt_shape)
    return filt


if config.cudnn_enabled:  # pragma: no cover
    from cupy import cudnn

    def _complex(func, data1, data2, *kargs, **kwargs):
        """Helper function to convert func to support complex floats."""
        xp = backend.get_array_module(data1)
        data1r = xp.real(data1)
        data1i = xp.imag(data1)
        data2r = xp.real(data2)
        data2i = xp.imag(data2)

        outputr = func(data1r, data2r, *kargs, **kwargs)
        outputr -= func(data1i, data2i, *kargs, **kwargs)
        outputi = func(data1i, data2r, *kargs, **kwargs)
        outputi += func(data1r, data2i, *kargs, **kwargs)

        output = outputr + 1j * outputi
        output = output.astype(data1.dtype, copy=False)
        return output

    def _convolve_cuda(
        data, filt, mode="full", strides=None, multi_channel=False
    ):
        xp = backend.get_array_module(data)

        D, b, B, m, n, s, c_i, c_o, p = _get_convolve_params(
            data.shape, filt.shape, mode, strides, multi_channel
        )

        if D == 1:
            return _convolve_cuda(
                xp.expand_dims(data, -1),
                xp.expand_dims(filt, -1),
                mode=mode,
                strides=list(strides) + [1] if strides is not None else None,
                multi_channel=multi_channel,
            ).squeeze(-1)
        elif D > 3:
            raise ValueError(
                f"cuDNN convolution only supports 1, 2, or 3D, got {D}."
            )

        dilations = (1,) * D
        groups = 1
        auto_tune = True
        tensor_core = "auto"
        if mode == "full":
            pads = tuple(n_d - 1 for n_d in n)
        elif mode == "valid":
            pads = (0,) * D

        data = data.reshape((B, c_i) + m)
        filt = filt.reshape((c_o, c_i) + n)
        output = xp.empty((B, c_o) + p, dtype=data.dtype)
        filt = util.flip(filt, axes=range(-D, 0))
        cudnn.convolution_forward(
            data,
            filt,
            None,
            output,
            pads,
            s,
            dilations,
            groups,
            auto_tune=auto_tune,
            tensor_core=tensor_core,
        )

        # Reshape.
        if multi_channel:
            output = output.reshape(b + (c_o,) + p)
        else:
            output = output.reshape(b + p)

        return output

    def _convolve_data_adjoint_cuda(
        output,
        filt,
        data_shape,
        mode="full",
        strides=None,
        multi_channel=False,
    ):
        xp = backend.get_array_module(output)

        D, b, B, m, n, s, c_i, c_o, p = _get_convolve_params(
            data_shape, filt.shape, mode, strides, multi_channel
        )

        if D == 1:
            return _convolve_data_adjoint_cuda(
                xp.expand_dims(output, -1),
                xp.expand_dims(filt, -1),
                list(data_shape) + [1],
                mode=mode,
                strides=list(strides) + [1] if strides is not None else None,
                multi_channel=multi_channel,
            ).squeeze(-1)
        elif D > 3:
            raise ValueError(
                f"cuDNN convolution only supports 1, 2 or 3D, got {D}."
            )

        dilations = (1,) * D
        groups = 1
        auto_tune = True
        tensor_core = "auto"
        deterministic = False
        if mode == "full":
            pads = tuple(n_d - 1 for n_d in n)
        elif mode == "valid":
            pads = (0,) * D

        output = output.reshape((B, c_o) + p)
        filt = filt.reshape((c_o, c_i) + n)
        data = xp.empty((B, c_i) + m, dtype=output.dtype)
        filt = util.flip(filt, axes=range(-D, 0))
        cudnn.convolution_backward_data(
            filt,
            output,
            None,
            data,
            pads,
            s,
            dilations,
            groups,
            deterministic=deterministic,
            auto_tune=auto_tune,
            tensor_core=tensor_core,
        )

        # Reshape.
        data = data.reshape(data_shape)

        return data

    def _convolve_filter_adjoint_cuda(
        output,
        data,
        filt_shape,
        mode="full",
        strides=None,
        multi_channel=False,
    ):
        xp = backend.get_array_module(data)

        D, b, B, m, n, s, c_i, c_o, p = _get_convolve_params(
            data.shape, filt_shape, mode, strides, multi_channel
        )
        if D == 1:
            return _convolve_filter_adjoint_cuda(
                xp.expand_dims(output, -1),
                xp.expand_dims(data, -1),
                list(filt_shape) + [1],
                mode=mode,
                strides=list(strides) + [1] if strides is not None else None,
                multi_channel=multi_channel,
            ).squeeze(-1)
        elif D > 3:
            raise ValueError(
                f"cuDNN convolution only supports 1, 2 or 3D, got {D}."
            )

        dilations = (1,) * D
        groups = 1
        auto_tune = True
        tensor_core = "auto"
        deterministic = False
        if mode == "full":
            pads = tuple(n_d - 1 for n_d in n)
        elif mode == "valid":
            pads = (0,) * D

        data = data.reshape((B, c_i) + m)
        output = output.reshape((B, c_o) + p)
        filt = xp.empty((c_o, c_i) + n, dtype=output.dtype)
        cudnn.convolution_backward_filter(
            data,
            output,
            filt,
            pads,
            s,
            dilations,
            groups,
            deterministic=deterministic,
            auto_tune=auto_tune,
            tensor_core=tensor_core,
        )
        filt = util.flip(filt, axes=range(-D, 0))
        filt = filt.reshape(filt_shape)

        return filt



import unittest

import numpy as np
import numpy.testing as npt

from sigpy import alg, linop

if __name__ == "__main__":
    unittest.main()


class TestAlg(unittest.TestCase):
    def Ax_setup(self, n):
        A = np.eye(n) + 0.1 * np.ones([n, n])
        x = np.arange(n, dtype=np.float32)
        return A, x

    def Ax_y_setup(self, n, lamda):
        A, x = self.Ax_setup(n)
        y = A @ x
        x_numpy = np.linalg.solve(A.T @ A + lamda * np.eye(n), A.T @ y)

        return A, x_numpy, y

    def test_PowerMethod(self):
        n = 5
        A, x = self.Ax_setup(n)
        x_hat = np.random.random([n, 1])
        alg_method = alg.PowerMethod(lambda x: A.T @ A @ x, x_hat)
        while not alg_method.done():
            alg_method.update()

        s_numpy = np.linalg.svd(A, compute_uv=False)[0]
        s_sigpy = np.linalg.norm(A @ x_hat)
        npt.assert_allclose(s_numpy, s_sigpy, atol=1e-3)

    def test_GradientMethod(self):
        n = 5
        lamda = 0.1
        A, x_numpy, y = self.Ax_y_setup(n, lamda)

        # Compute step-size
        lipschitz = np.linalg.svd(
            A.T @ A + lamda * np.eye(n), compute_uv=False
        )[0]
        alpha = 1.0 / lipschitz

        for accelerate in [True, False]:
            for proxg in [None, lambda alpha, x: x / (1 + lamda * alpha)]:
                with self.subTest(accelerate=accelerate, proxg=proxg):
                    x_sigpy = np.zeros([n])

                    def gradf(x):
                        gradf_x = A.T @ (A @ x - y)
                        if proxg is None:
                            gradf_x += lamda * x

                        return gradf_x

                    alg_method = alg.GradientMethod(
                        gradf,
                        x_sigpy,
                        alpha,
                        accelerate=accelerate,
                        proxg=proxg,
                        max_iter=1000,
                    )

                    while not alg_method.done():
                        alg_method.update()

                    npt.assert_allclose(x_sigpy, x_numpy)

    def test_ConjugateGradient(self):
        n = 5
        lamda = 0.1
        A, x_numpy, y = self.Ax_y_setup(n, lamda)
        x = np.zeros([n])
        alg_method = alg.ConjugateGradient(
            lambda x: A.T @ A @ x + lamda * x, A.T @ y, x, max_iter=1000
        )
        while not alg_method.done():
            alg_method.update()

        npt.assert_allclose(x, x_numpy)

    def test_PrimalDualHybridGradient(self):
        n = 5
        lamda = 0.1
        A, x_numpy, y = self.Ax_y_setup(n, lamda)

        # Compute step-size
        lipschitz = np.linalg.svd(np.matmul(A.T, A), compute_uv=False)[0]
        tau = 1.0 / lipschitz
        sigma = 1.0

        x = np.zeros([n])
        u = np.zeros([n])
        alg_method = alg.PrimalDualHybridGradient(
            lambda alpha, u: (u - alpha * y) / (1 + alpha),
            lambda alpha, x: x / (1 + lamda * alpha),
            lambda x: A @ x,
            lambda x: A.T @ x,
            x,
            u,
            tau,
            sigma,
            max_iter=1000,
        )
        while not alg_method.done():
            alg_method.update()

        npt.assert_allclose(x, x_numpy)

    def test_AugmentedLagrangianMethod(self):
        n = 5
        lamda = 0.1
        A, x_numpy, y = self.Ax_y_setup(n, lamda)

        # Solve 1 / 2 \| A x - y \|_2^2 + lamda * \| z \|_2^2 s.t. x = z
        mu = 1
        x_z = np.zeros([2 * n])
        v = np.zeros([n])

        def minL():
            x = x_z[:n]
            z = x_z[n:]
            x[:] = np.linalg.solve(
                A.T @ A + mu * np.eye(n), A.T @ y - v + mu * z
            )
            z[:] = (mu * x + v) / (mu + lamda)

        def h(x_z):
            x = x_z[:n]
            z = x_z[n:]
            return x - z

        alg_method = alg.AugmentedLagrangianMethod(
            minL, None, h, x_z, None, v, mu
        )
        while not alg_method.done():
            alg_method.update()

        x = x_z[:n]
        npt.assert_allclose(x, x_numpy)

    def test_NewtonsMethod(self):
        n = 5
        lamda = 0.1
        A, x_numpy, y = self.Ax_y_setup(n, lamda)

        def gradf(x):
            gradf_x = A.T @ (A @ x - y)
            gradf_x += lamda * x

            return gradf_x

        def inv_hessf(x):
            Id = np.eye(n)
            return lambda x: np.linalg.pinv(A.T @ A + lamda * Id) @ x

        for beta in [1, 0.5]:
            with self.subTest(beta=beta):
                if beta < 1:

                    def f(x):
                        f_x = 1 / 2 * np.linalg.norm(A @ x - y) ** 2
                        f_x += lamda / 2 * np.linalg.norm(x) ** 2

                        return f_x

                else:
                    f = None

                x = np.zeros(n)
                alg_method = alg.NewtonsMethod(
                    gradf, inv_hessf, x, beta=beta, f=f
                )
                while not alg_method.done():
                    alg_method.update()

                npt.assert_allclose(x, x_numpy)

    def test_GerchbergSaxton(self):
        n = 10
        lamda = 0.1
        A, x_numpy, y = self.Ax_y_setup(n, lamda)
        y = np.expand_dims(np.csingle(y), 1)
        x_numpy = np.expand_dims(x_numpy, 1)
        A = np.csingle(A)
        A = linop.MatMul(y.shape, A)
        x0 = np.zeros(A.ishape, dtype=complex)

        alg_method = alg.GerchbergSaxton(
            A, y, x0, max_iter=100, tol=10e-9, lamb=lamda
        )

        while not alg_method.done():
            alg_method.update()

        phs = np.conj(x_numpy * alg_method.x / abs(x_numpy * alg_method.x))
        npt.assert_allclose(alg_method.x * phs, x_numpy, rtol=1e-6)

    def test_SDMM(self):
        n = 5
        lamda = 0.1
        A, x_numpy, y = self.Ax_y_setup(n, lamda)
        y = np.expand_dims(y, 1)
        A = linop.MatMul(np.expand_dims(x_numpy, 1).shape, A)

        c_norm = None
        c_max = None
        mu = 10**8  # big mu ok since no constraints used
        rho_norm = 1
        rho_max = 1
        lam = 0.1
        cg_iters = 5
        max_iter = 10

        L = []
        c = [1]
        rho = [1]
        for ii in range(len(y) - 1):
            c.append(0.00012**2)
            rho.append(0.001)

        alg_method = alg.SDMM(
            A,
            y,
            lam,
            L=L,
            c=c,
            c_max=c_max,
            c_norm=c_norm,
            mu=mu,
            rho=rho,
            rho_max=rho_max,
            rho_norm=rho_norm,
            eps_pri=10**-5,
            eps_dual=10**-2,
            max_cg_iter=cg_iters,
            max_iter=max_iter,
        )

        while not alg_method.done():
            alg_method.update()

        npt.assert_allclose(np.squeeze(abs(alg_method.x)), x_numpy)



# -*- coding: utf-8 -*-
"""This module provides an abstract class Alg for iterative algorithms,
and implements commonly used methods, such as gradient methods,
Newton's method, and the augmented Lagrangian method.
"""
import numpy as np

import sigpy as sp
from sigpy import backend, util


class Alg(object):
    """Abstraction for iterative algorithms.

    The standard way of using an :class:`Alg` object, say alg, is as follows:

    >>> while not alg.done():
    >>>     alg.update()

    The user is free to run other things in the while loop.
    An :class:`Alg` object is meant to run once.
    Once done, the object should not be run again.

    When creating a new :class:`Alg` class, the user should supply
    an _update() function
    to perform the iterative update, and optionally a _done() function
    to determine when to terminate the iteration. The default _done() function
    simply checks whether the number of iterations has reached the maximum.

    The interface for each :class:`Alg` class should not depend on
    Linop or Prox explicitly.
    For example, if the user wants to design an
    :class:`Alg` class to accept a Linop, say A,
    as an argument, then it should also accept any function that can be called
    to compute x -> A(x). Similarly, to accept a Prox, say proxg,
    as an argument,
    the Alg class should accept any function that can be called to compute
    alpha, x -> proxg(x).

    Args:
        max_iter (int): Maximum number of iterations.

    Attributes:
        max_iter (int): Maximum number of iterations.
        iter (int): Current iteration.

    """

    def __init__(self, max_iter):
        self.max_iter = max_iter
        self.iter = 0

    def _update(self):
        raise NotImplementedError

    def _done(self):
        return self.iter >= self.max_iter

    def update(self):
        """Perform one update step.

        Call the user-defined _update() function and increment iter.
        """
        self._update()
        self.iter += 1

    def done(self):
        """Return whether the algorithm is done.

        Call the user-defined _done() function.
        """
        return self._done()


class PowerMethod(Alg):
    """Power method to estimate maximum eigenvalue and eigenvector.

    Args:
        A (Linop or function): Function to a hermitian linear mapping.
        x (array): Variable to optimize over.
        max_iter (int): Maximum number of iterations.

    Attributes:
        max_eig (float): Maximum eigenvalue of `A`.

    """

    def __init__(self, A, x, norm_func=None, max_iter=30):
        self.A = A
        self.x = x
        self.max_eig = np.infty
        self.norm_func = norm_func
        super().__init__(max_iter)

    def _update(self):
        y = self.A(self.x)
        device = backend.get_device(y)
        xp = device.xp
        with device:
            if self.norm_func is None:
                self.max_eig = xp.linalg.norm(y).item()
            else:
                self.max_eig = self.norm_func(y)

            backend.copyto(self.x, y / self.max_eig)

    def _done(self):
        return self.iter >= self.max_iter


class GradientMethod(Alg):
    r"""First order gradient method.

    For the simplest setting when proxg is not specified,
    the method considers the objective function:

    .. math:: \min_x f(x)

    where :math:`f` is (sub)-differentiable and performs the update:

    .. math:: x_\text{new} = x - \alpha \nabla f(x)

    When proxg is specified, the method considers the composite
    objective function:

    .. math:: f(x) + g(x)

    where :math:`f` is (sub)-differentiable and :math:`g` is simple,
    and performs the update:

    .. math:: x_\text{new} = \text{prox}_{\alpha g}(x - \alpha \nabla f(x))

    Nesterov's acceleration is supported by toggling the `accelerate`
    input option.

    Args:
        gradf (function): function to compute :math:`\nabla f`.
        x (array): variable to optimize over.
        alpha (float or None): step size, or initial step size
             if backtracking line-search is on.
        proxg (Prox, function or None): Prox or function to compute
            proximal operator of :math:`g`.
        accelerate (bool): toggle Nesterov acceleration.
        max_iter (int): maximum number of iterations.
        tol (float): Tolerance for stopping condition.

    References:
        Nesterov, Y. E. (1983).
        A method for solving the convex programming problem
        with convergence rate O (1/k^ 2).
        In Dokl. Akad. Nauk SSSR (Vol. 269, pp. 543-547).

        Beck, A., & Teboulle, M. (2009).
        A fast iterative shrinkage-thresholding algorithm
        for linear inverse problems.
        SIAM journal on imaging sciences, 2(1), 183-202.

    """

    def __init__(
        self,
        gradf,
        x,
        alpha,
        proxg=None,
        accelerate=False,
        max_iter=100,
        tol=0,
    ):
        self.gradf = gradf
        self.alpha = alpha
        self.accelerate = accelerate
        self.proxg = proxg
        self.x = x
        self.tol = tol

        self.device = backend.get_device(x)
        with self.device:
            if self.accelerate:
                self.z = self.x.copy()
                self.t = 1

        self.resid = np.infty
        super().__init__(max_iter)

    def _update(self):
        xp = self.device.xp
        with self.device:
            x_old = self.x.copy()

            if self.accelerate:
                backend.copyto(self.x, self.z)

            # Perform update
            util.axpy(self.x, -self.alpha, self.gradf(self.x))
            if self.proxg is not None:
                backend.copyto(self.x, self.proxg(self.alpha, self.x))

            if self.accelerate:
                t_old = self.t
                self.t = (1 + (1 + 4 * t_old**2) ** 0.5) / 2
                backend.copyto(
                    self.z, self.x + ((t_old - 1) / self.t) * (self.x - x_old)
                )

            self.resid = xp.linalg.norm(self.x - x_old).item() / self.alpha

    def _done(self):
        return (self.iter >= self.max_iter) or self.resid <= self.tol


class ConjugateGradient(Alg):
    r"""Conjugate gradient method.

    Solves for:

    .. math:: A x = b

    where A is a Hermitian linear operator.

    Args:
        A (Linop or function): Linop or function to compute A.
        b (array): Observation.
        x (array): Variable.
        P (function or None): Preconditioner.
        max_iter (int): Maximum number of iterations.
        tol (float): Tolerance for stopping condition.

    """

    def __init__(self, A, b, x, P=None, max_iter=100, tol=0):
        self.A = A
        self.b = b
        self.P = P
        self.x = x
        self.tol = tol
        self.device = backend.get_device(x)
        with self.device:
            xp = self.device.xp
            self.r = b - self.A(self.x)

            if self.P is None:
                z = self.r
            else:
                z = self.P(self.r)

            if max_iter > 1:
                self.p = z.copy()
            else:
                self.p = z

            self.not_positive_definite = False
            self.rzold = xp.real(xp.vdot(self.r, z))
            self.resid = self.rzold.item() ** 0.5

        super().__init__(max_iter)

    def _update(self):
        with self.device:
            xp = self.device.xp
            Ap = self.A(self.p)
            pAp = xp.real(xp.vdot(self.p, Ap)).item()
            if pAp <= 0:
                self.not_positive_definite = True
                return

            self.alpha = self.rzold / pAp
            util.axpy(self.x, self.alpha, self.p)
            if self.iter < self.max_iter - 1:
                util.axpy(self.r, -self.alpha, Ap)
                if self.P is not None:
                    z = self.P(self.r)
                else:
                    z = self.r

                rznew = xp.real(xp.vdot(self.r, z))
                beta = rznew / self.rzold
                util.xpay(self.p, beta, z)
                self.rzold = rznew

            self.resid = self.rzold.item() ** 0.5

    def _done(self):
        return (
            self.iter >= self.max_iter
            or self.not_positive_definite
            or self.resid <= self.tol
        )


class PrimalDualHybridGradient(Alg):
    r"""Primal dual hybrid gradient.

    Considers the problem:

    .. math:: \min_x \max_u - f^*(u) + g(x) + \left<Ax, u\right>

    Or equivalently:

    .. math:: \min_x f(A x) + g(x)

    where f, and g are simple.

    Args:
        proxfc (function): Function to compute proximal operator of f^*.
        proxg (function): Function to compute proximal operator of g.
        A (function): Function to compute a linear mapping.
        AH (function): Function to compute the adjoint linear mapping of `A`.
        x (array): Primal solution.
        u (array): Dual solution.
        tau (float or array): Primal step-size.
        sigma (float or array): Dual step-size.
        gamma_primal (float): Strong convexity parameter of g.
        gamma_dual (float): Strong convexity parameter of f^*.
        max_iter (int): Maximum number of iterations.
        tol (float): Tolerance for stopping condition.

    References:
       Chambolle, A., & Pock, T. (2011).
       A first-order primal-dual algorithm for convex problems with
       applications to imaging.
       Journal of mathematical imaging and vision, 40(1), 120-145.

    """

    def __init__(
        self,
        proxfc,
        proxg,
        A,
        AH,
        x,
        u,
        tau,
        sigma,
        theta=1,
        gamma_primal=0,
        gamma_dual=0,
        max_iter=100,
        tol=0,
    ):
        self.proxfc = proxfc
        self.proxg = proxg
        self.tol = tol

        self.A = A
        self.AH = AH

        self.u = u
        self.x = x

        self.tau = tau
        self.sigma = sigma
        self.theta = theta
        self.gamma_primal = gamma_primal
        self.gamma_dual = gamma_dual

        self.x_device = backend.get_device(x)
        self.u_device = backend.get_device(u)

        with self.x_device:
            self.x_ext = self.x.copy()

        if self.gamma_primal > 0:
            xp = self.x_device.xp
            with self.x_device:
                self.tau_min = xp.amin(xp.abs(tau)).item()

        if self.gamma_dual > 0:
            xp = self.u_device.xp
            with self.u_device:
                self.sigma_min = xp.amin(xp.abs(sigma)).item()

        self.resid = np.infty

        super().__init__(max_iter)

    def _update(self):
        # Update dual.
        util.axpy(self.u, self.sigma, self.A(self.x_ext))
        backend.copyto(self.u, self.proxfc(self.sigma, self.u))

        # Update primal.
        with self.x_device:
            x_old = self.x.copy()
            util.axpy(self.x, -self.tau, self.AH(self.u))
            backend.copyto(self.x, self.proxg(self.tau, self.x))

        # Update step-size if neccessary.
        if self.gamma_primal > 0 and self.gamma_dual == 0:
            with self.x_device:
                xp = self.x_device.xp
                theta = 1 / (1 + 2 * self.gamma_primal * self.tau_min) ** 0.5
                self.tau *= theta
                self.tau_min *= theta

            with self.u_device:
                self.sigma /= theta
        elif self.gamma_primal == 0 and self.gamma_dual > 0:
            with self.u_device:
                xp = self.u_device.xp
                theta = 1 / (1 + 2 * self.gamma_dual * self.sigma_min) ** 0.5
                self.sigma *= theta
                self.sigma_min *= theta

            with self.x_device:
                self.tau /= theta
        else:
            theta = self.theta

        # Extrapolate primal.
        with self.x_device:
            xp = self.x_device.xp
            x_diff = self.x - x_old
            self.resid = xp.linalg.norm(x_diff / self.tau**0.5).item()
            backend.copyto(self.x_ext, self.x + theta * x_diff)

    def _done(self):
        return (self.iter >= self.max_iter) or (self.resid <= self.tol)


class AltMin(Alg):
    """Alternating minimization.

    Args:
        min1 (function): Function to minimize over variable 1.
        min2 (function): Funciton to minimize over variable 2.
        max_iter (int): Maximum number of iterations.

    """

    def __init__(self, min1, min2, max_iter=30):
        self.min1 = min1
        self.min2 = min2
        super().__init__(max_iter)

    def _update(self):
        self.min1()
        self.min2()


class AugmentedLagrangianMethod(Alg):
    r"""Augmented Lagrangian method for constrained optimization.

    Consider the equality and inequality constrained problem:

    .. math:: \min_{x: g(x) \leq 0, h(x) = 0} f(x)

    And perform the following update steps:

    .. math::
        x \in \text{argmin}_{x} L(x, u, v, \mu)\\
        u = [u + \mu g(x)]_+
        v = v + \mu h(x)

    where :math:`L(x, u, v, \mu)`: is the augmented Lagrangian function:

    .. math::
        L(x, u, v, \mu) = f(x) + \frac{\mu}{2}(
        \|[g(x) + \frac{u}{\mu}]_+\|_2^2 + \|h(x) + \frac{v}{\mu}\|_2^2)

    Args:
        minL (function): a function that minimizes the augmented Lagrangian.
        g (None or function): a function that takes :math:`x` as input,
            and outputs :math:`g(x)`, the inequality constraints.
        h (None or function): a function that takes :math:`x` as input,
            and outputs :math:`h(x)`, the equality constraints.
        x (array): primal variable.
        u (array): dual variable for inequality constraints.
        v (array): dual variable for equality constraints.
        mu (scalar): step size.
        max_iter (int): maximum number of iterations.

    """

    def __init__(self, minL, g, h, x, u, v, mu, max_iter=30):
        self.minL = minL
        self.g = g
        self.h = h
        self.x = x
        self.u = u
        self.v = v
        self.mu = mu
        super().__init__(max_iter)

    def _update(self):
        self.minL()
        if self.g is not None:
            device = backend.get_device(self.u)
            xp = device.xp
            with device:
                util.axpy(self.u, self.mu, self.g(self.x))
                backend.copyto(self.u, xp.clip(self.u, 0, np.infty))

        if self.h is not None:
            util.axpy(self.v, self.mu, self.h(self.x))


class ADMM(Alg):
    r"""Alternating Direction Method of Multipliers.

    Consider the equality constrained problem:

    .. math:: \min_{x: A x + B z = c} f(x) + g(z)

    And perform the following update steps:

    .. math::
        x = \text{argmin}_{x} L_\mu(x, z, u)\\
        z = \text{argmin}_{z} L_\mu(x, z, u)\\
        u = u + A x + B z - c

    where :math:`L(x, u, v, \mu)`: is the augmented Lagrangian function:

    .. math::
        L_\rho(x, z, u) = f(x) + g(z) + \frac{\rho}{2}\|A x + B z - c + u\|_2^2

    Args:
        minL_x (function): a function that minimizes L w.r.t. x.
        minL_z (function): a function that minimizes L w.r.t. z.
        x (array): primal variable 1.
        z (array): primal variable 2.
        u (array): scaled dual variable.
        max_iter (int): maximum number of iterations.

    """

    def __init__(self, minL_x, minL_z, x, z, u, A, B, c, max_iter=30):
        self.minL_x = minL_x
        self.minL_z = minL_z
        self.x = x
        self.z = z
        self.u = u
        self.A = A
        self.B = B
        self.c = c
        super().__init__(max_iter)

    def _update(self):
        self.minL_x()
        self.minL_z()
        self.u += self.A(self.x) + self.B(self.z) - self.c


class SDMM(Alg):
    r"""Simultaneous Direction Method of Multipliers. Can be used for
    unconstrained or constrained optimization with several constraints.

    Solves the problem of form:

    .. math::

        \min_{x} \frac{1}{2}\left\|Ax-d\right\|_2^2
        + \frac{\lambda}{2}\left\|x\right\|_2^2

        s.t. \left\|L_{i}x\right\|_2^2 < c_{i}

    In SDMM, constraints are typically specified as a list of L linear
    operators and c constraints. This algorithm gives the user the option to
    provide either a list of L's and c's or a single maximum (c_max) and/or
    norm constraint (c_norm) with implicit L.

    Solution variable x can be found in alg_method.x.

    Args:
        A (Linop): a system matrix Linop.
        d (array): observation.
        lam (float): scaling parameter.
        L (list of arrays): list of constraint linear operator arrays. If not
            used, provide empty list {}.
        c (list of floats): list of constraints, constraining the L2 norm of
            :math:`L_{i}x`
        mu (float): proximal scaling factor.
        rho (list of floats): list of L scaling parameters, which each one
            corresponding to one L constraint linear array.
        rho_max (float): max constraint scaling parameter (if c_max provided).
        rho_norm (float): norm constraint scaling parameter (if c_norm
            provided).
        eps_pri (float): primal variable error tolerance.
        eps_dual (float): dual variable error tolerance.
        c_max (float): maximum value constraint.
        c_norm (float): norm constraint.
        max_cg_iter (int): maximum number of unconstrained CG iterations per
            SDMM iteration.
        max_iter (int): maximum number of SDMM iterations.

    References:
        Moolekamp, F. and Melchior, P. (2017). 'Block-Simultaneous Direction
        Method of Multipliers: A proximal primal-dual splitting algorithm for
        nonconvex problems with multiple constraints.' arXiv.

    """

    def __init__(
        self,
        A,
        d,
        lam,
        L,
        c,
        mu,
        rho,
        rho_max,
        rho_norm,
        eps_pri=10**-5,
        eps_dual=10**-2,
        c_max=None,
        c_norm=None,
        max_cg_iter=30,
        max_iter=1000,
    ):
        self.A = A
        self.d = d
        self.lam = lam
        self.L = L
        self.c = c
        self.mu = mu
        self.rho = rho
        self.rho_max = rho_max
        self.rho_norm = rho_norm
        self.eps_pri = eps_pri
        self.eps_dual = eps_dual
        self.c_max = c_max
        self.c_norm = c_norm
        self.max_cg_iter = max_cg_iter
        self.stop = False  # stop criterion collector variable
        self.device = backend.get_device(d)
        super().__init__(max_iter)

        M = len(self.L)
        with self.device:
            xp = self.device.xp
            self.x = xp.zeros(self.A.ishape, dtype=complex).flatten()
            self.x = xp.expand_dims(self.x, axis=1)
            self.z, self.u = [], []

            for ii in range(M):
                self.z.append(L[ii] @ self.x)
                self.u.append(
                    xp.expand_dims(
                        xp.zeros(xp.shape(L[ii])[0], dtype=xp.complex), axis=1
                    )
                )
            if c_max is not None:
                self.zMax = self.x
                self.uMax = xp.zeros(xp.shape(self.x), dtype=xp.complex)
            if c_norm is not None:
                self.zNorm = self.x
                self.uNorm = xp.zeros(xp.shape(self.x), dtype=xp.complex)

    def prox_rhog(self, v, c):
        with self.device:
            xp = self.device.xp
            if xp.real(xp.linalg.norm(v) ** 2) > c:
                z = v * xp.sqrt(c) / xp.sqrt(xp.real(xp.linalg.norm(v)))
            else:
                z = v
            return z

    def prox_rhog_max(self, v, c):
        with self.device:
            xp = self.device.xp
            z = v
            indices = xp.where((abs(z) ** 2 > c))
            z[indices] = xp.sqrt(c) * z[indices] / xp.absolute(z[indices])
            return z

    def prox_muf(self, v, mu, A, x, d, lam, nCGiters):
        with self.device:
            xp = self.device.xp
            d = xp.vstack((d, xp.sqrt(1 / mu) * v, xp.sqrt(lam) * x))
            Am = self.Amult(x, A, mu, lam)
            int_method = ConjugateGradient(
                Am.H * Am, Am.H * d, x, max_iter=nCGiters
            )

            while not int_method.done():
                int_method.update()

            return int_method.x

    def Amult(self, x, A, mu, lam):
        M = sp.linop.Multiply(x.shape, np.ones(x.shape) * np.sqrt(1 / mu))
        L = sp.linop.Multiply(x.shape, np.ones(x.shape) * np.sqrt(lam))
        Y = sp.linop.Vstack((A, M, L))
        Ry = sp.linop.Reshape((Y.oshape[0], 1), Y.oshape)
        Y = Ry * Y
        return Y

    def _update(self):
        with self.device:
            xp = self.device.xp
            # evaluate objective
            v = self.x
            for ii in range(len(self.L)):
                v -= (
                    self.mu
                    / self.rho[ii]
                    * xp.transpose(self.L[ii])
                    @ (self.L[ii] @ self.x - self.z[ii] + self.u[ii])
                )
            if self.c_max is not None:
                x_min_z_pl_u = self.x - self.zMax + self.uMax
                v -= self.mu / self.rho_max * x_min_z_pl_u
            if self.c_norm is not None:
                x_min_z_pl_u = self.x - self.zNorm + self.uNorm
                v -= self.mu / self.rho_norm * x_min_z_pl_u

            self.x = self.prox_muf(
                v, self.mu, self.A, self.x, self.d, self.lam, self.max_cg_iter
            )

            # run through constraints
            z_old = self.z
            for ii in range(len(self.L)):
                self.z[ii] = self.prox_rhog(
                    self.L[ii] @ (self.x + self.u[ii]), self.c[ii]
                )
                self.u[ii] += self.L[ii] @ self.x - self.z[ii]

            if self.c_max is not None:
                zMax_old = self.zMax
                self.zMax = self.prox_rhog_max(self.x + self.uMax, self.c_max)
                self.uMax = self.uMax + self.x - self.zMax
            if self.c_norm is not None:
                zNorm_old = self.zNorm
                self.zNorm = self.prox_rhog(self.x + self.uNorm, self.c_norm)
                self.uNorm += self.x - self.zNorm

            # check the stopping criteria
            self.stop = True
            rMax, sMax = 0, 0
            for ii in range(len(self.L)):
                # primal residual
                r = self.L[ii] @ self.x - self.z[ii]
                # dual residual
                dz = self.z[ii] - z_old[ii]
                s = 1 / self.rho[ii] * xp.transpose(self.L[ii]) * dz
                if (
                    xp.linalg.norm(r) > self.eps_pri
                    or xp.linalg.norm(s) > self.eps_dual
                ):
                    self.stop = False
                if xp.linalg.norm(r) > rMax:
                    rMax = xp.linalg.norm(r)
                if xp.linalg.norm(s) > sMax:
                    sMax = xp.linalg.norm(s)
            if self.c_norm is not None:
                r = self.x - self.zNorm
                s = 1 / self.rho_norm * (self.zNorm - zNorm_old)
                if (
                    xp.linalg.norm(r) > self.eps_pri
                    or xp.linalg.norm(s) > self.eps_dual
                ):
                    self.stop = False
                if xp.linalg.norm(r) > rMax:
                    rMax = xp.linalg.norm(r)
                if xp.linalg.norm(s) > sMax:
                    sMax = xp.linalg.norm(s)
            if self.c_max is not None:
                r = self.x - self.zMax
                s = 1 / self.rho_max * (self.zMax - zMax_old)
                if (
                    xp.linalg.norm(r) > self.eps_pri
                    or xp.linalg.norm(s) > self.eps_dual
                ):
                    self.stop = False
                if xp.linalg.norm(r) > rMax:
                    rMax = xp.linalg.norm(r)
                if xp.linalg.norm(s) > sMax:
                    sMax = xp.linalg.norm(s)

    def _done(self):
        return self.iter >= self.max_iter or self.stop


class NewtonsMethod(Alg):
    """Newton's Method.

    Args:
        gradf (function) - A function gradf(x): x -> gradient of f at x.
        inv_hessf (function) - A function H(x): x -> inverse Hessian of f at x,
            which is another function: y -> inverse Hessian of f at x times y.
        x (function) - solution.
        beta (scalar): backtracking linesearch factor.
             Enables backtracking when beta < 1.
        f (function or None): function to compute :math:`f`
             for backtracking line-search.
        max_iter (int): maximum number of iterations.
        tol (float): Tolerance for stopping condition.

    """

    def __init__(
        self, gradf, inv_hessf, x, beta=1, f=None, max_iter=10, tol=0
    ):
        if beta < 1 and f is None:
            raise TypeError(
                "Cannot do backtracking linesearch without specifying f."
            )

        self.gradf = gradf
        self.inv_hessf = inv_hessf
        self.x = x
        self.lamda = np.infty
        self.beta = beta
        self.f = f
        self.residual = np.infty
        self.tol = tol

        super().__init__(max_iter)

    def _update(self):
        device = backend.get_device(self.x)
        xp = device.xp
        with device:
            gradf_x = self.gradf(self.x)
            p = -self.inv_hessf(self.x)(gradf_x)
            self.lamda2 = -xp.real(xp.vdot(p, gradf_x)).item()
            if self.lamda2 < 0:
                raise ValueError(
                    "Direction is not descending. Got lamda2={}. "
                    "inv_hessf might not be defined correctly.".format(
                        self.lamda2
                    )
                )

            x_new = self.x + p
            if self.beta < 1:
                fx = self.f(self.x)
                alpha = 1
                while self.f(x_new) > fx - alpha / 2 * self.lamda2:
                    alpha *= self.beta
                    x_new = self.x + alpha * p

            backend.copyto(self.x, x_new)
            self.residual = self.lamda2**0.5

    def _done(self):
        return self.iter >= self.max_iter or self.residual <= self.tol


class GerchbergSaxton(Alg):
    """Gerchberg-Saxton method, also called the variable exchange method.
    Iterative method for recovery of a signal from the amplitude of linear
    measurements |Ax|.

    Args:
        A (Linop): system matrix Linop.
        y (array): observations.
        max_iter (int): maximum number of iterations.
        tol (float): optimization stopping tolerance.
        lamb (float): Tikhonov regularization value.

    """

    def __init__(self, A, y, x0, max_iter=500, tol=0, max_tol=0, lamb=0):
        self.A = A
        self.Aholder = A
        self.y = y
        self.x = x0
        self.max_iter = max_iter
        self.iter = 0
        self.tol = tol
        self.max_tol = max_tol
        self.lamb = lamb
        self.residual = np.infty

    def _update(self):
        device = backend.get_device(self.y)
        xp = device.xp
        with device:
            y_hat = self.y * xp.exp(1j * xp.angle(self.A * self.x))
            system = self.A.H * self.A + self.lamb * sp.linop.Identity(
                self.A.ishape
            )
            b = self.A.H * y_hat

            alg_internal = ConjugateGradient(system, b, self.x, max_iter=5)

            while not alg_internal.done():
                alg_internal.update()
                self.x = alg_internal.x

        self.residual = xp.sum(
            xp.absolute(xp.absolute(self.A * self.x) - self.y)
        )
        self.iter += 1

    def _done(self):
        over_iter = self.iter >= self.max_iter
        under_tol = self.residual <= self.tol
        return over_iter or under_tol



import unittest

import numpy as np
import numpy.testing as npt

import sigpy.mri.rf as rf

if __name__ == "__main__":
    unittest.main()


class TestB1sel(unittest.TestCase):
    def test_b1sel_generic(self):
        dt = 2e-6  # sampling period
        d1 = 0.01  # passband ripple
        d2 = 0.01  # stopband ripple
        tb = 4  # time-bandwidth product
        ptype = "ex"  # 'st', 'ex', 'inv' or 'sat'
        pbw = 0.5  # gauss, passband width
        pbc = 5  # gauss, passband center
        flip = np.pi / 4  # radians, flip angle

        [rf_am, rf_fm] = rf.b1sel.dz_b1_rf(
            dt, tb, ptype, flip, pbw, pbc, d1, d2
        )
        b1 = np.arange(
            0, 2 * pbc, 2 * pbc / np.size(rf_am) * 4
        )  # b1 grid we simulate the pulse over
        b1 = np.reshape(b1, (np.size(b1), 1))
        [a, b] = rf.sim.abrm_nd(
            2 * np.pi * dt * rf_fm,
            b1,
            2 * np.pi * 4258 * dt * np.reshape(rf_am, (np.size(rf_am), 1)),
        )
        mxy = -2 * np.real(a * b) + 1j * np.imag(np.conj(a) ** 2 - b**2)

        pts = np.array([mxy[10], mxy[int(len(b1) / 2)], mxy[len(b1) - 10]])
        npt.assert_almost_equal(abs(pts), np.array([0, 0.7, 0]), decimal=1)

    def test_b1sel_gslider(self):
        g = 5
        flip = np.pi / 2
        ptype = "ex"  # 'ex' or 'st'
        tb = 12
        d1 = 0.01
        d2 = 0.01
        pbc = 1  # gauss, passband center
        pbw = 0.25  # passband width
        dt = 2e-6  # seconds, sampling rate
        [om1, dom] = rf.b1sel.dz_b1_gslider_rf(
            dt, g, tb, ptype, flip, pbw, pbc, d1, d2
        )

        n = np.shape(om1)[0]
        b1 = np.arange(
            0, 2 * pbc, 2 * pbc / n * 4
        )  # b1 grid we simulate the pulse over
        b1 = np.reshape(b1, (np.size(b1), 1))
        [a, b] = rf.sim.abrm_nd(
            2 * np.pi * dt * dom[:, 0],
            b1,
            2 * np.pi * 4258 * dt * np.reshape(om1[:, 0], (n, 1)),
        )
        mxy = -2 * np.real(a * b) + 1j * np.imag(np.conj(a) ** 2 - b**2)

        pts = np.array([mxy[10], mxy[int(len(b1) / 2)], mxy[len(b1) - 10]])
        npt.assert_almost_equal(abs(pts), np.array([0, 1, 0]), decimal=2)



import unittest

import numpy as np
import numpy.testing as npt

import sigpy as sp
import sigpy.mri.rf as rf

if __name__ == "__main__":
    unittest.main()


class TestSlr(unittest.TestCase):
    def test_st(self):
        #  check to make sure profile roughly matches anticipated within d1, d2
        N = 128
        tb = 16
        filts = ["ls", "ms", "pm", "min", "max"]
        for idx, filt in enumerate(filts):
            pulse = sp.mri.rf.dzrf(
                N, tb, ptype="st", ftype=filt, d1=0.01, d2=0.01
            )

            m = np.abs(sp.fft(pulse, norm=None))

            pts = np.array(
                [m[int(N / 2 - 10)], m[int(N / 2)], m[int(N / 2 + 10)]]
            )
            npt.assert_almost_equal(pts, np.array([0, 1, 0]), decimal=2)

    def test_inv(self):
        #  also provides testing of sim. Check inv profile.
        tb = 8
        N = 128
        d1 = 0.01
        d2 = 0.01
        ptype = "ex"
        filts = ["min", "max"]  # filts produce inconsistent inversions

        for idx, filt in enumerate(filts):
            pulse = rf.slr.dzrf(N, tb, ptype, filt, d1, d2)

            [_, b] = rf.sim.abrm(pulse, np.arange(-2 * tb, 2 * tb, 0.01))
            mz = 1 - 2 * np.abs(b) ** 2

            pts = np.array(
                [
                    mz[int(len(mz) / 2 - len(mz) / 3)],
                    mz[int(len(mz) / 2)],
                    mz[int(len(mz) / 2 + len(mz) / 3)],
                ]
            )

            npt.assert_almost_equal(pts, np.array([1, -0.2, 1]), decimal=1)

    def test_root_flipped(self):
        tb = 12
        N = 128
        d1 = 0.01
        d2 = 0.001
        flip = np.pi / 2
        ptype = "ex"
        [bsf, d1, d2] = rf.slr.calc_ripples(ptype, d1, d2)
        b = bsf * rf.slr.dzmp(N, tb, d1, d2)
        b = b[::-1]
        [pulse, _] = rf.slr.root_flip(b, d1, flip, tb, verbose=False)

        [_, b] = rf.sim.abrm(pulse, np.arange(-2 * tb, 2 * tb, 0.01))
        mz = 1 - 2 * np.abs(b) ** 2

        pts = np.array(
            [
                mz[int(len(mz) / 2 - len(mz) / 3)],
                mz[int(len(mz) / 2)],
                mz[int(len(mz) / 2 + len(mz) / 3)],
            ]
        )

        npt.assert_almost_equal(pts, np.array([1, 0.2, 1]), decimal=1)

    def test_recursive(self):
        # Design the pulses
        nseg = 3  # number of EPI segments/RF Pulses
        tb = 4
        n = 200
        se_seq = True
        tb_ref = 8  # time-bandwidth of ref pulse
        [pulses, _] = rf.slr.dz_recursive_rf(nseg, tb, n, se_seq, tb_ref)

        mz = np.ones(np.size(np.arange(-4 * tb, 4 * tb, 0.01)))
        for ii in range(0, nseg):
            [a, b] = rf.sim.abrm(
                pulses[:, ii], np.arange(-4 * tb, 4 * tb, 0.01), True
            )
            mxy = 2 * mz * np.multiply(np.conj(a), b)
            mz = mz * (1 - 2 * np.abs(b) ** 2)

            pts = np.array(
                [
                    mxy[int(len(mxy) / 2 - len(mxy) / 3)],
                    mxy[int(len(mxy) / 2)],
                    mxy[int(len(mxy) / 2 + len(mxy) / 3)],
                ]
            )

            npt.assert_almost_equal(abs(pts), np.array([0, 0.5, 0]), decimal=1)

    def test_gslider(self):
        n = 512
        g = 5
        ex_flip = 90 * np.pi / 180
        tb = 12
        d1 = 0.01
        d2 = 0.01
        phi = np.pi

        pulses = rf.slr.dz_gslider_rf(
            n, g, ex_flip, phi, tb, d1, d2, cancel_alpha_phs=True
        )

        for gind in range(1, g + 1):
            [a, b] = rf.sim.abrm(
                pulses[:, gind - 1], np.arange(-2 * tb, 2 * tb, 0.01), True
            )
            mxy = 2 * np.multiply(np.conj(a), b)

            pts = np.array(
                [
                    mxy[int(len(mxy) / 2 - len(mxy) / 3)],
                    mxy[int(len(mxy) / 2)],
                    mxy[int(len(mxy) / 2 + len(mxy) / 3)],
                ]
            )

            npt.assert_almost_equal(abs(pts), np.array([0, 1, 0]), decimal=2)



# -*- coding: utf-8 -*-
"""MRI utilities.
"""
import numpy as np

import sigpy as sp

__all__ = ["get_cov", "whiten", "tseg_off_res_b_ct", "apply_tseg"]


def get_cov(noise):
    """Get covariance matrix from noise measurements.

    Args:
        noise (array): Noise measurements of shape [num_coils, ...]

    Returns:
        array: num_coils x num_coils covariance matrix.

    """
    num_coils = noise.shape[0]
    X = noise.reshape([num_coils, -1])
    X -= np.mean(X, axis=-1, keepdims=True)
    cov = np.matmul(X, X.T.conjugate())

    return cov


def whiten(ksp, cov):
    """Whitens k-space measurements.

    Args:
        ksp (array): k-space measurements of shape [num_coils, ...]
        cov (array): num_coils x num_coils covariance matrix.

    Returns:
        array: whitened k-space array.

    """
    num_coils = ksp.shape[0]

    x = ksp.reshape([num_coils, -1])

    L = np.linalg.cholesky(cov)
    x_w = np.linalg.solve(L, x)
    ksp_w = x_w.reshape(ksp.shape)

    return ksp_w


def tseg_off_res_b_ct(b0, bins, lseg, dt, T):
    """Creates B and Ct matrices needed for time-segmented off-resonance
    compensation.

    Args:
        b0 (array): inhomogeneity matrix.
        bins (int): number of histogram bins to use.
        lseg (int): number of time segments.
        dt (float): hardware dwell time (ms).
        T (float): length of pulse (ms).

    Returns:
        2-element tuple containing

        - **B** (*array*): temporal interpolator.
        - **Ct** (*array*): off-resonance phase at each time segment center.
    """

    # create time vector
    t = np.linspace(0, T, int(T / dt))
    hist_wt, bin_edges = np.histogram(
        np.imag(2j * np.pi * np.concatenate(b0)), bins
    )

    # Build B and Ct
    bin_centers = bin_edges[1:] - bin_edges[1] / 2
    zk = 0 + 1j * bin_centers
    tl = np.linspace(0, lseg, lseg) / lseg * T / 1000  # time seg centers
    # calculate off-resonance phase @ each time seg, for hist bins
    ch = np.exp(-np.expand_dims(tl, axis=1) @ np.expand_dims(zk, axis=0))
    w = np.diag(np.sqrt(hist_wt))
    p = np.linalg.pinv(w @ np.transpose(ch)) @ w
    b = p @ np.exp(
        -np.expand_dims(zk, axis=1) @ np.expand_dims(t, axis=0) / 1000
    )
    b = np.transpose(b)
    b0_v = np.expand_dims(2j * np.pi * np.concatenate(b0), axis=0)
    ct = np.transpose(np.exp(-np.expand_dims(tl, axis=1) @ b0_v))

    return b, ct


def apply_tseg(array_in, coord, b, ct, fwd=True):
    """Apply the temporal interpolator and phase shift maps calculated

    Args:
        array_in (array): array to apply correction to.
        coord (array): coordinates for noncartesian trajectories. [Nt 2].
        b (array): temporal interpolator.
        ct (array): off-resonance phase at each time segment center.
        fwd (Boolean): indicates forward direction (img -> kspace) or
            backward (kspace->img)

    Returns:
        out (array): array with correction applied.
    """

    # get number of time segments from B input.
    lseg = b.shape[1]
    dim = array_in.shape[0]

    out = 0
    if fwd:
        for ii in range(lseg):
            ctd = np.reshape(ct[:, ii] * array_in.flatten(), (dim, dim))
            out = out + b[:, ii] * sp.fourier.nufft(ctd, coord * 20)

    else:
        for ii in range(lseg):
            ctd = np.reshape(
                np.conj(ct[:, ii]) * array_in.flatten(), (dim, dim)
            )
            out = out + sp.fourier.nufft(ctd, coord * 20) * np.conj(b[:, ii])

    return np.expand_dims(out, 1)



# -*- coding: utf-8 -*-
"""MRI RF excitation pulse design functions,
    including SLR and small tip spatial design
"""

import numpy as np
from scipy.interpolate import interp1d

import sigpy as sp
from sigpy import backend
from sigpy.mri import rf as rf

__all__ = ["stspa", "stspk"]


def stspa(
    target,
    sens,
    coord,
    dt,
    roi=None,
    alpha=0,
    b0=None,
    tseg=None,
    st=None,
    phase_update_interval=float("inf"),
    explicit=False,
    max_iter=1000,
    tol=1e-6,
):
    """Small tip spatial domain method for multicoil parallel excitation.
       Allows for constrained or unconstrained designs.

    Args:
        target (array): desired magnetization profile. [dim dim]
        sens (array): sensitivity maps. [Nc dim dim]
        coord (array): coordinates for noncartesian trajectories. [Nt 2]
        dt (float): hardware sampling dwell time.
        roi (array): array for error weighting, specify spatial ROI. [dim dim]
        alpha (float): regularization term, if unconstrained.
        b0 (array): B0 inhomogeneity map [dim dim]. For explicit matrix
            building.
        tseg (None or Dictionary): parameters for time-segmented off-resonance
            correction. Parameters are 'b0' (array), 'dt' (float),
            'lseg' (int), and 'n_bins' (int). Lseg is the number of
            time segments used, and n_bins is the number of histogram bins.
        st (None or Dictionary): 'subject to' constraint parameters. Parameters
            are avg power 'cNorm' (float), peak power 'cMax' (float),
            'mu' (float), 'rhoNorm' (float), 'rhoMax' (float), 'cgiter' (int),
            'max_iter' (int), 'L' (list of arrays), 'c' (float), 'rho' (float),
            and 'lam' (float). These parameters are explained in detail in the
            SDMM documentation.
        phase_update_interval (int): number of iters between exclusive phase
            updates. If 0, no phase updates performed.
        explicit (bool): Use explicit matrix.
        max_iter (int): max number of iterations.
        tol (float): allowable error.

    Returns:
        array: pulses out.

    References:
        Grissom, W., Yip, C., Zhang, Z., Stenger, V. A., Fessler, J. A.
        & Noll, D. C.(2006).
        Spatial Domain Method for the Design of RF Pulses in Multicoil
        Parallel Excitation. Magnetic resonance in medicine, 56, 620-629.
    """
    Nc = sens.shape[0]
    Nt = coord.shape[0]
    device = backend.get_device(target)
    xp = device.xp
    with device:
        pulses = xp.zeros((Nc, Nt), xp.complex64)

        # set up the system matrix
        if explicit:
            A = rf.linop.PtxSpatialExplicit(sens, coord, dt, target.shape, b0)
        else:
            A = sp.mri.linop.Sense(
                sens, coord, weights=None, tseg=tseg, ishape=target.shape
            ).H

        # handle the Ns * Ns error weighting ROI matrix
        W = sp.linop.Multiply(A.oshape, xp.ones(target.shape))
        if roi is not None:
            W = sp.linop.Multiply(A.oshape, roi)

        # apply ROI
        A = W * A

        # Unconstrained, use conjugate gradient
        if st is None:
            Id = sp.linop.Identity((Nc, coord.shape[0]))
            b = A.H * W * target

            alg_method = sp.alg.ConjugateGradient(
                A.H * A + alpha * Id,
                b,
                pulses,
                P=None,
                max_iter=max_iter,
                tol=tol,
            )

        # Constrained case, use SDMM
        else:
            # vectorize target for SDMM
            target = W * target
            d = xp.expand_dims(target.flatten(), axis=0)
            alg_method = sp.alg.SDMM(
                A,
                d,
                st["lam"],
                st["L"],
                st["c"],
                st["mu"],
                st["rho"],
                st["rhoMax"],
                st["rhoNorm"],
                10**-5,
                10**-2,
                st["cMax"],
                st["cNorm"],
                st["cgiter"],
                st["max_iter"],
            )

        # perform the design: apply optimization method to find solution pulse
        while not alg_method.done():
            # phase_update switch
            if (alg_method.iter > 0) and (
                alg_method.iter % phase_update_interval == 0
            ):
                target = xp.abs(target) * xp.exp(
                    1j * xp.angle(xp.reshape(A * alg_method.x, target.shape))
                )
                b = A.H * target
                alg_method.b = b

            alg_method.update()

        if st is not None:
            pulses = xp.reshape(alg_method.x, [Nc, Nt])
        return pulses


def stspk(
    mask,
    sens,
    n_spokes,
    fov,
    dx_max,
    gts,
    sl_thick,
    tbw,
    dgdtmax,
    gmax,
    alpha=1,
    iter_dif=0.01,
):
    """Small tip spokes parallel transmit pulse designer.

    Args:
       mask (ndarray): region in which to optimize flip angle uniformity
           in slice. [dim dim]
       sens (ndarray): sensitivity maps. [nc dim dim]
       n_spokes (int): number of spokes to be created in the design.
       fov (float): excitation FOV (cm).
       dx_max (float): max. resolution of the trajectory (cm).
       gts (float): hardware sampling dwell time (s).
       sl_thick (float): slice thickness (mm).
       tbw (int): time-bandwidth product.
       dgdtmax (float): max gradient slew (g/cm/s).
       gmax (float): max gradient amplitude (g/cm).
       alpha (float): regularization parameter.
       iter_dif (float): for each spoke, the difference in cost btwn.
          successive iterations at which to terminate MLS iterations.

    Returns:
       2-element tuple containing

        - **pulses** (*array*): RF waveform out.
        - **g** (*array*): corresponding gradient, in g/cm.

    References:
       Grissom, W., Khalighi, M., Sacolick, L., Rutt, B. & Vogel, M (2012).
       Small-tip-angle spokes pulse design using interleaved greedy and
       local optimization methods. Magnetic Resonance in Medicine, 68(5),
       1553-62.
    """

    device = backend.get_device(sens)
    xp = device.xp
    with device:
        nc = sens.shape[0]

        kmax = 1 / dx_max  # /cm, max spatial freq of trajectory
        # greedy kx, ky grid
        kxs, kys = xp.meshgrid(
            xp.linspace(-kmax / 2, kmax / 2 - 1 / fov, int(fov * kmax)),
            xp.linspace(-kmax / 2, kmax / 2 - 1 / fov, int(fov * kmax)),
        )
        # vectorize the grid
        kxs = kxs.flatten()
        kys = kys.flatten()

        # remove DC
        dc = xp.intersect1d(xp.where((kxs == 0)), xp.where((kys == 0)))[0]
        kxs = xp.concatenate([kxs[:dc], kxs[dc + 1 :]])
        kys = xp.concatenate([kys[:dc], kys[dc + 1 :]])

        # step 2: design the weights
        # initial kx/ky location is DC
        k = xp.expand_dims(xp.array([0, 0]), 0)

        # initial target phase
        phs = xp.zeros((xp.count_nonzero(mask), 1), dtype=xp.complex64)

        for ii in range(n_spokes):
            # build Afull (and take only 0 locations into matrix)
            Anum = rf.PtxSpatialExplicit(
                sens, k, gts, mask.shape, ret_array=True
            )
            Anum = Anum[~(Anum == 0).all(1)]

            # design wfull using MLS:
            # initialize wfull
            sys_a = Anum.conj().T @ Anum + alpha * xp.eye((ii + 1) * nc)
            sys_b = Anum.conj().T @ xp.exp(1j * phs)
            w_full = xp.linalg.solve(sys_a, sys_b)

            err = Anum @ w_full - xp.exp(1j * phs)
            cost = err.conj().T @ err + alpha * w_full.conj().T @ w_full
            cost = xp.real(cost)
            cost_old = 10 * cost  # to get the loop going
            while xp.absolute(cost - cost_old) > iter_dif * cost_old:
                cost_old = cost
                phs = xp.angle(Anum @ w_full)
                w_full = xp.linalg.solve(
                    (Anum.conj().T @ Anum + alpha * xp.eye((ii + 1) * nc)),
                    (Anum.conj().T @ xp.exp(1j * phs)),
                )
                err = Anum @ w_full - xp.exp(1j * phs)
                cost = xp.real(
                    err.conj().T @ err + alpha * w_full.conj().T @ w_full
                )

            # add a spoke using greedy method
            if ii < n_spokes - 1:
                r = xp.exp(1j * phs) - Anum @ w_full
                rfnorm = xp.zeros(kxs.shape, dtype=xp.complex64)
                for jj in range(kxs.size):
                    ks_test = xp.expand_dims(xp.array([kxs[jj], kys[jj]]), 0)
                    Anum = rf.PtxSpatialExplicit(
                        sens, ks_test, gts, mask.shape, ret_array=True
                    )
                    Anum = Anum[~(Anum == 0).all(1)]

                    rfm = xp.linalg.solve(
                        (Anum.conj().T @ Anum), (Anum.conj().T @ r)
                    )
                    rfnorm[jj] = xp.linalg.norm(rfm)

                ind = xp.argmax(rfnorm)
                k_new = xp.expand_dims(xp.array([kxs[ind], kys[ind]]), 0)

                if ii % 2 != 0:  # add to end of pulse
                    k = xp.concatenate((k, k_new))
                else:  # add to beginning of pulse
                    k = xp.concatenate((k_new, k))

                # remove chosen point from candidates
                kxs = xp.concatenate([kxs[:ind], kxs[ind + 1 :]])
                kys = xp.concatenate([kys[:ind], kys[ind + 1 :]])

        # from our spoke selections, build the whole waveforms

        # first, design our gradient waveforms:
        g = rf.spokes_grad(k, tbw, sl_thick, gmax, dgdtmax, gts)

        # design our rf
        # calc. the size of the traps in our gz waveform- will use to calc rf
        area = tbw / (sl_thick / 10) / 4257  # thick*kwid=twb, kwid=gam*area
        [subgz, nramp] = rf.min_trap_grad(area, gmax, dgdtmax, gts)
        npts = 128
        subrf = rf.dzrf(npts, tbw, "st")

        n_plat = subgz.size - 2 * nramp  # time points on trap plateau
        # interpolate to stretch out waveform to appropriate length
        f = interp1d(
            np.arange(0, npts, 1) / npts, subrf, fill_value="extrapolate"
        )
        subrf = f(xp.arange(0, n_plat, 1) / n_plat)
        subrf = xp.concatenate((xp.zeros(nramp), subrf, xp.zeros(nramp)))

        pulses = xp.kron(xp.reshape(w_full, (nc, n_spokes)), subrf)

        # add zeros for gzref
        rf_ref = xp.zeros((nc, g.shape[1] - pulses.shape[1]))
        pulses = xp.concatenate((pulses, rf_ref), 1)

        return pulses, g



# -*- coding: utf-8 -*-
"""This module contains an abstraction class Prox for proximal operators,
and provides commonly used proximal operators, including soft-thresholding,
l1 ball projection, and box constraints.
"""
import numpy as np

from sigpy import backend, thresh, util


class Prox(object):
    r"""Abstraction for proximal operator.

    Prox can be called on a float (:math:`\alpha`) and
    a NumPy or CuPy array (:math:`x`) to perform a proximal operation.

    .. math::
        \text{prox}_{\alpha g} (y) =
        \text{argmin}_x \frac{1}{2} || x - y ||_2^2 + \alpha g(x)

    Prox can be stacked, and conjugated.

    Args:
        shape: Input/output shape.
        repr_str (string or None): default: class name.

    Attributes:
        shape: Input/output shape.

    """

    def __init__(self, shape, repr_str=None):
        self.shape = list(shape)

        if repr_str is None:
            self.repr_str = self.__class__.__name__
        else:
            self.repr_str = repr_str

    def _check_shape(self, input):
        for i1, i2 in zip(input.shape, self.shape):
            if i2 != -1 and i1 != i2:
                raise ValueError(
                    "shape mismatch for {s}, got {input_shape}.".format(
                        s=self, input_shape=input.shape
                    )
                )

    def __call__(self, alpha, input):
        try:
            self._check_shape(input)
            output = self._prox(alpha, input)
            self._check_shape(output)
        except Exception as e:
            raise RuntimeError("Exceptions from {}.".format(self)) from e

        return output

    def __repr__(self):
        return "<{shape} {repr_str} Prox>.".format(
            shape=self.shape, repr_str=self.repr_str
        )


class Conj(Prox):
    r"""Returns the proximal operator for the convex conjugate function.

    The proximal operator of the convex conjugate function
    :math:`g^*` is defined as:

    .. math::
        \text{prox}_{\alpha g^*} (x) =
        x - \alpha \text{prox}_{\frac{1}{\alpha} g} (\frac{1}{\alpha} x)

    """

    def __init__(self, prox):
        self.prox = prox
        super().__init__(prox.shape)

    def _prox(self, alpha, input):
        with backend.get_device(input):
            return input - alpha * self.prox(1 / alpha, input / alpha)


class NoOp(Prox):
    r"""Proximal operator for empty function. Equivalant to an identity function.

    Args:
       shape (tuple of ints): Input shape

    """

    def __init__(self, shape):
        super().__init__(shape)

    def _prox(self, alpha, input):
        return input


class Stack(Prox):
    r"""Stack outputs of proximal operators.

    Args:
       proxs (list of proxs): Prox of the same shape.

    """

    def __init__(self, proxs):
        self.nops = len(proxs)
        assert self.nops > 0

        self.proxs = proxs
        self.shapes = [prox.shape for prox in proxs]
        shape = [sum(util.prod(prox.shape) for prox in proxs)]

        super().__init__(shape)

    def _prox(self, alpha, input):
        with backend.get_device(input):
            if np.isscalar(alpha):
                alphas = [alpha] * self.nops
            else:
                alphas = util.split(alpha, self.shapes)

            inputs = util.split(input, self.shapes)
            outputs = [
                prox(alpha, input)
                for prox, input, alpha in zip(self.proxs, inputs, alphas)
            ]
            output = util.vec(outputs)

            return output


class UnitaryTransform(Prox):
    r"""Unitary transform input space.

    Returns a proximal operator that does

    .. math::
        A^H \text{prox}_{\alpha g}(A x)

    Args:
        prox (Prox): Proximal operator.
        A (Linop): Unitary linear operator.

    """

    def __init__(self, prox, A):
        self.prox = prox
        self.A = A

        super().__init__(A.ishape)

    def _prox(self, alpha, input):
        return self.A.H(self.prox(alpha, self.A(input)))


class L2Reg(Prox):
    r"""Proximal operator for l2 regularization.

    .. math::
        \min_x \frac{1}{2} \|x - y\|_2^2 + \frac{\lambda}{2}\|x-z\|_2^2 + h(x)

    Args:
        shape (tuple of ints): Input shape.
        lamda (float): Regularization parameter.
        y (scalar or array): Bias term.
        proxh (Prox): optional additional proximal operator.

    """

    def __init__(self, shape, lamda, y=None, proxh=None):
        self.lamda = lamda
        self.y = y
        self.proxh = proxh

        super().__init__(shape)

    def _prox(self, alpha, input):
        with backend.get_device(input):
            output = input.copy()
            if self.y is not None:
                output += (self.lamda * alpha) * self.y

            output /= 1 + self.lamda * alpha

            if self.proxh is not None:
                return self.proxh(alpha / (1 + self.lamda * alpha), output)

        return output


class L2Proj(Prox):
    r"""Proximal operator for l2 norm projection.

    .. math::
        \min_x \frac{1}{2} \| x - y \|_2^2 + 1\{\| x \|_2 < \epsilon\}

    Args:
        shape (tuple of ints): Input shape.
        epsilon (float): Regularization parameter.
        y (scalar or array): Bias term.

    """

    def __init__(self, shape, epsilon, y=0, axes=None):
        self.epsilon = epsilon
        self.y = y
        self.axes = axes

        super().__init__(shape)

    def _prox(self, alpha, input):
        with backend.get_device(input):
            return (
                thresh.l2_proj(self.epsilon, input - self.y, self.axes)
                + self.y
            )


class LInfProj(Prox):
    r"""Proximal operator for l-infinity ball projection.

    .. math::
        \min_x \frac{1}{2} \| x - y \|_2^2 + 1\{\| x \|_\infty < \epsilon\}

    Args:
        shape (tuple of ints): Input shape.
        epsilon (float): Regularization parameter.
        y (scalar or array): Bias term.

    """

    def __init__(self, shape, epsilon, bias=None, axes=None):
        self.epsilon = epsilon
        self.bias = bias
        self.axes = axes

        super().__init__(shape)

    def _prox(self, alpha, input):
        with backend.get_device(input):
            return thresh.linf_proj(self.epsilon, input, bias=self.bias)


class PsdProj(Prox):
    r"""Proximal operator for positive semi-definite matrices.

    .. math::
        \min_x \frac{1}{2} \| X - Y \|_2^2 + 1\{\| X \succeq 0\}

    Args:
        shape (tuple of ints): Input shape.

    """

    def _prox(self, alpha, input):
        with backend.get_device(input):
            return thresh.psd_proj(input)


class L1Reg(Prox):
    r"""Proximal operator for l1 regularization.

    .. math::
        \min_x \frac{1}{2} \| x - y \|_2^2 + \lambda \| x \|_1

    Args:
        shape (tuple of ints): input shape
        lamda (float): regularization parameter

    """

    def __init__(self, shape, lamda):
        self.lamda = lamda

        super().__init__(shape)

    def _prox(self, alpha, input):
        with backend.get_device(input):
            return thresh.soft_thresh(self.lamda * alpha, input)


class L1Proj(Prox):
    r"""Proximal operator for l1 norm projection.

    .. math::
        \min_x \frac{1}{2} \| x - y \|_2^2 + 1\{\| x \|_1 < \epsilon\}

    Args:
        shape (tuple of ints): input shape.
        epsilon (float): regularization parameter.

    """

    def __init__(self, shape, epsilon):
        self.epsilon = epsilon

        super().__init__(shape)

    def _prox(self, alpha, input):
        with backend.get_device(input):
            return thresh.l1_proj(self.epsilon, input)


class BoxConstraint(Prox):
    r"""Box constraint proximal operator.

    .. math::
        \min_{x : l \leq x \leq u} \frac{1}{2} \| x - y \|_2^2

    Args:
        shape (tuple of ints): input shape.
        lower (scalar or array): lower limit.
        upper (scalar or array): upper limit.

    """

    def __init__(self, shape, lower, upper):
        self.lower = lower
        self.upper = upper
        super().__init__(shape)

    def _prox(self, alpha, input):
        device = backend.get_device(input)
        xp = device.xp

        with device:
            return xp.clip(input, self.lower, self.upper)



import unittest

import numpy as np
import numpy.testing as npt

from sigpy import backend, config, linop, pytorch

if config.pytorch_enabled:
    import torch

    if __name__ == "__main__":
        unittest.main()

    devices = [backend.cpu_device]
    if config.cupy_enabled:
        devices.append(backend.Device(0))

    class TestPytorch(unittest.TestCase):
        def test_to_pytorch(self):
            for dtype in [np.float32, np.float64]:
                for device in devices:
                    with self.subTest(device=device, dtype=dtype):
                        xp = device.xp
                        array = xp.array([1, 2, 3], dtype=dtype)
                        tensor = pytorch.to_pytorch(array)
                        array[0] = 0
                        torch.testing.assert_allclose(
                            tensor,
                            torch.tensor(
                                [0, 2, 3],
                                dtype=tensor.dtype,
                                device=tensor.device,
                            ),
                        )

        def test_to_pytorch_complex(self):
            for dtype in [np.complex64, np.complex128]:
                for device in devices:
                    with self.subTest(device=device, dtype=dtype):
                        xp = device.xp
                        array = xp.array([1 + 1j, 2 + 2j, 3 + 3j], dtype=dtype)
                        tensor = pytorch.to_pytorch(array)
                        array[0] = 0
                        torch.testing.assert_allclose(
                            tensor,
                            torch.tensor(
                                [[0, 0], [2, 2], [3, 3]],
                                dtype=tensor.dtype,
                                device=tensor.device,
                            ),
                        )

        def test_from_pytorch(self):
            for dtype in [torch.float32, torch.float64]:
                for device in devices:
                    with self.subTest(device=device, dtype=dtype):
                        if device == backend.cpu_device:
                            torch_device = torch.device("cpu")
                        else:
                            torch_device = torch.device("cuda:0")

                        tensor = torch.tensor(
                            [1, 2, 3], dtype=dtype, device=torch_device
                        )
                        array = pytorch.from_pytorch(tensor)
                        array[0] = 0
                        np.testing.assert_allclose(
                            tensor.cpu().numpy(), [0, 2, 3]
                        )

        def test_from_pytorch_complex(self):
            for dtype in [torch.float32, torch.float64]:
                for device in devices:
                    with self.subTest(device=device, dtype=dtype):
                        if device == backend.cpu_device:
                            torch_device = torch.device("cpu")
                        else:
                            torch_device = torch.device("cuda:0")

                        tensor = torch.tensor(
                            [[1, 1], [2, 2], [3, 3]],
                            dtype=dtype,
                            device=torch_device,
                        )
                        array = pytorch.from_pytorch(tensor, iscomplex=True)
                        xp = device.xp
                        xp.testing.assert_array_equal(
                            array, [1 + 1j, 2 + 2j, 3 + 3j]
                        )
                        array[0] -= 1
                        np.testing.assert_allclose(
                            tensor.cpu().numpy(), [[0, 1], [2, 2], [3, 3]]
                        )

        def test_to_pytorch_function(self):
            A = linop.Resize([5], [3])
            x = np.array([1, 2, 3], float)
            y = np.ones([5])

            with self.subTest("forward"):
                f = pytorch.to_pytorch_function(A).apply
                x_torch = pytorch.to_pytorch(x)
                npt.assert_allclose(f(x_torch).detach().numpy(), A(x))

            with self.subTest("adjoint"):
                y_torch = pytorch.to_pytorch(y)
                loss = (f(x_torch) - y_torch).pow(2).sum() / 2
                loss.backward()
                npt.assert_allclose(
                    x_torch.grad.detach().numpy(), A.H(A(x) - y)
                )

        def test_to_pytorch_function_complex(self):
            A = linop.FFT([3])
            x = np.array([1 + 1j, 2 + 2j, 3 + 3j], complex)
            y = np.ones([3], complex)

            with self.subTest("forward"):
                f = pytorch.to_pytorch_function(
                    A, input_iscomplex=True, output_iscomplex=True
                ).apply
                x_torch = pytorch.to_pytorch(x)
                npt.assert_allclose(
                    f(x_torch).detach().numpy().ravel(), A(x).view(float)
                )

            with self.subTest("adjoint"):
                y_torch = pytorch.to_pytorch(y)
                loss = (f(x_torch) - y_torch).pow(2).sum() / 2
                loss.backward()
                npt.assert_allclose(
                    x_torch.grad.detach().numpy().ravel(),
                    A.H(A(x) - y).view(float),
                )



import unittest

import numpy as np
import numpy.testing as npt

from sigpy import wavelet

if __name__ == "__main__":
    unittest.main()


class TestWavelet(unittest.TestCase):
    def test_fwt(self):
        n = 8
        input = np.zeros(n, dtype=np.float32)
        input[0] = 1
        npt.assert_allclose(
            wavelet.fwt(input, level=1, wave_name="haar"),
            [1 / 2**0.5, 0, 0, 0, 1 / 2**0.5, 0, 0, 0],
        )

    def test_fwt_iwt(self):
        for n in range(5, 11):
            input = np.zeros(n, dtype=np.float32)
            input[0] = 1
            _, coeff_slices = wavelet.get_wavelet_shape([n])
            npt.assert_allclose(
                wavelet.iwt(wavelet.fwt(input), [n], coeff_slices), input
            )



# -*- coding: utf-8 -*-
"""Configuration.

This module contains flags to turn on and off optional modules.

"""
import warnings
from importlib import util

numba_enabled = util.find_spec("numba") is not None
if numba_enabled:
    try:
        import numba  # noqa
    except ImportError as e:
        warnings.warn(
            f"Importing numba failed. "
            f"For more details, see the error stack below:\n{e}"
        )
        numba_enabled = False

pywt_enabled = util.find_spec("pywt") is not None
if pywt_enabled:
    try:
        import pywt  # noqa
    except ImportError as e:
        warnings.warn(
            f"Importing pywt failed. "
            f"For more details, see the error stack below:\n{e}"
        )
        pywt_enabled = False

cupy_enabled = util.find_spec("cupy") is not None
if cupy_enabled:
    try:
        import cupy  # noqa
    except ImportError as e:
        warnings.warn(
            f"Importing cupy failed. "
            f"For more details, see the error stack below:\n{e}"
        )
        cupy_enabled = False

if cupy_enabled:  # pragma: no cover
    try:
        cudnn_enabled = util.find_spec("cupy.cuda.cudnn") is not None
        if cudnn_enabled:
            from cupy import cudnn  # noqa: F401
    except ImportError as e:
        warnings.warn(
            f"Importing cupy.cuda.cudnn failed. "
            f"For more details, see the error stack below:\n{e}"
        )
        cudnn_enabled = False
    try:
        nccl_enabled = util.find_spec("cupy.cuda.nccl") is not None
        if nccl_enabled:
            from cupy.cuda import nccl  # noqa: F401
    except ImportError as e:
        warnings.warn(
            f"Importing cupy.cuda.nccl failed. "
            f"For more details, see the error stack below:\n{e}"
        )
        nccl_enabled = False
else:
    cudnn_enabled = False
    nccl_enabled = False

mpi4py_enabled = util.find_spec("mpi4py") is not None

# This is to catch an import error when the cudnn in cupy (system) and pytorch
# (built in) are in conflict.
if util.find_spec("torch") is not None:
    try:
        import torch  # noqa

        pytorch_enabled = True
    except ImportError:
        print("Warning : Pytorch installed but can import")
        pytorch_enabled = False
else:
    pytorch_enabled = False



import unittest

import numpy as np
import numpy.testing as npt

from sigpy.learn import util

if __name__ == "__main__":
    unittest.main()


class TestUtil(unittest.TestCase):
    def test_labels_to_scores(self):
        labels = np.array([0, 1, 2])

        scores = util.labels_to_scores(labels)

        npt.assert_allclose(scores, [[1, 0, 0], [0, 1, 0], [0, 0, 1]])

    def test_scores_to_labels(self):
        scores = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])

        labels = util.scores_to_labels(scores)

        npt.assert_allclose(labels, [0, 1, 2])



import unittest

import numpy as np

from sigpy import block, config

if config.cupy_enabled:
    import cupy as cp

if __name__ == "__main__":
    unittest.main()


class TestInterp(unittest.TestCase):
    def test_array_to_blocks(self):
        xps = [np]
        if config.cupy_enabled:
            xps.append(cp)

        for xp in xps:
            for dtype in [np.float32, np.complex64]:
                for ndim in [1, 2, 3]:
                    with self.subTest(xp=xp, dtype=dtype, ndim=ndim):
                        input = xp.array(
                            [0, 1, 2, 3, 4, 5], dtype=dtype
                        ).reshape([6] + [1] * (ndim - 1))

                        blk_shape = [1] + [1] * (ndim - 1)
                        blk_strides = [1] + [1] * (ndim - 1)
                        output = xp.array(
                            [[0], [1], [2], [3], [4], [5]], dtype=dtype
                        ).reshape(
                            [6] + [1] * (ndim - 1) + [1] + [1] * (ndim - 1)
                        )
                        xp.testing.assert_allclose(
                            output,
                            block.array_to_blocks(
                                input, blk_shape, blk_strides
                            ),
                        )

                        blk_shape = [2] + [1] * (ndim - 1)
                        blk_strides = [1] + [1] * (ndim - 1)
                        output = xp.array(
                            [[0, 1], [1, 2], [2, 3], [3, 4], [4, 5]],
                            dtype=dtype,
                        ).reshape(
                            [5] + [1] * (ndim - 1) + [2] + [1] * (ndim - 1)
                        )
                        xp.testing.assert_allclose(
                            output,
                            block.array_to_blocks(
                                input, blk_shape, blk_strides
                            ),
                        )

                        blk_shape = [2] + [1] * (ndim - 1)
                        blk_strides = [2] + [1] * (ndim - 1)
                        output = xp.array(
                            [[0, 1], [2, 3], [4, 5]], dtype=dtype
                        ).reshape(
                            [3] + [1] * (ndim - 1) + [2] + [1] * (ndim - 1)
                        )
                        xp.testing.assert_allclose(
                            output,
                            block.array_to_blocks(
                                input, blk_shape, blk_strides
                            ),
                        )

                        blk_shape = [3] + [1] * (ndim - 1)
                        blk_strides = [2] + [1] * (ndim - 1)
                        output = xp.array(
                            [[0, 1, 2], [2, 3, 4]], dtype=dtype
                        ).reshape(
                            [2] + [1] * (ndim - 1) + [3] + [1] * (ndim - 1)
                        )
                        xp.testing.assert_allclose(
                            output,
                            block.array_to_blocks(
                                input, blk_shape, blk_strides
                            ),
                        )

    def test_blocks_to_array(self):
        xps = [np]
        if config.cupy_enabled:
            xps.append(cp)

        for xp in xps:
            for dtype in [np.float32, np.complex64]:
                for ndim in [1, 2, 3]:
                    with self.subTest(xp=xp, dtype=dtype, ndim=ndim):
                        shape = [6] + [1] * (ndim - 1)

                        blk_shape = [1] + [1] * (ndim - 1)
                        blk_strides = [1] + [1] * (ndim - 1)
                        input = xp.array(
                            [[0], [1], [2], [3], [4], [5]], dtype=dtype
                        ).reshape(
                            [6] + [1] * (ndim - 1) + [1] + [1] * (ndim - 1)
                        )
                        output = xp.array(
                            [0, 1, 2, 3, 4, 5], dtype=dtype
                        ).reshape([6] + [1] * (ndim - 1))
                        xp.testing.assert_allclose(
                            output,
                            block.blocks_to_array(
                                input, shape, blk_shape, blk_strides
                            ),
                        )

                        blk_shape = [2] + [1] * (ndim - 1)
                        blk_strides = [1] + [1] * (ndim - 1)
                        input = xp.array(
                            [[0, 1], [1, 2], [2, 3], [3, 4], [4, 5]],
                            dtype=dtype,
                        ).reshape(
                            [5] + [1] * (ndim - 1) + [2] + [1] * (ndim - 1)
                        )
                        output = xp.array(
                            [0, 2, 4, 6, 8, 5], dtype=dtype
                        ).reshape([6] + [1] * (ndim - 1))
                        xp.testing.assert_allclose(
                            output,
                            block.blocks_to_array(
                                input, shape, blk_shape, blk_strides
                            ),
                        )

                        blk_shape = [2] + [1] * (ndim - 1)
                        blk_strides = [2] + [1] * (ndim - 1)
                        input = xp.array(
                            [[0, 1], [2, 3], [4, 5]], dtype=dtype
                        ).reshape(
                            [3] + [1] * (ndim - 1) + [2] + [1] * (ndim - 1)
                        )
                        output = xp.array(
                            [0, 1, 2, 3, 4, 5], dtype=dtype
                        ).reshape([6] + [1] * (ndim - 1))
                        xp.testing.assert_allclose(
                            output,
                            block.blocks_to_array(
                                input, shape, blk_shape, blk_strides
                            ),
                        )

                        blk_shape = [3] + [1] * (ndim - 1)
                        blk_strides = [2] + [1] * (ndim - 1)
                        input = xp.array(
                            [[0, 1, 2], [2, 3, 4]], dtype=dtype
                        ).reshape(
                            [2] + [1] * (ndim - 1) + [3] + [1] * (ndim - 1)
                        )
                        output = xp.array(
                            [0, 1, 4, 3, 4, 0], dtype=dtype
                        ).reshape([6] + [1] * (ndim - 1))
                        xp.testing.assert_allclose(
                            output,
                            block.blocks_to_array(
                                input, shape, blk_shape, blk_strides
                            ),
                        )



# -*- coding: utf-8 -*-
"""Multiband RF Pulse Design functions.
"""
import numpy as np

__all__ = ["mb_phs_tab", "mb_rf", "dz_pins"]

from sigpy.mri.rf import slr as slr
from sigpy.mri.rf.trajgrad import trap_grad


def mb_rf(pulse_in, n_bands=3, band_sep=5, phs_0_pt="None"):
    r"""Multiband an input RF pulse.

    Args:
        pulse_in (array): samples of single-band RF pulse.
        n_bands (int): number of bands.
        band_sep (float): normalized slice separation.
        phs_0_pt (string): set of phases to use. Can be 'phs_mod' (Wong),
           'amp_mod' (Malik), 'quad_mod' (Grissom), or 'None'

    band_sep = slice_sep/slice_thick*tb, where tb is time-bandwidth product
    of the single-band pulse

    Returns:
        array: multibanded pulse out

    References:
        Wong, E. (2012). 'Optimized Phase Schedules for Minimizing Peak RF
        Power in Simultaneous Multi-Slice RF Excitation Pulses'. Proc. Intl.
        Soc. Mag. Reson. Med., 20 p. 2209.
        Malik, S. J., Price, A. N., and Hajnal, J. V. (2015). 'Optimized
        Amplitude Modulated Multi-Band RF pulses'. Proc. Intl. Soc. Mag.
        Reson. Med., 23 p. 2398.
    """

    if phs_0_pt != "None":
        phs = mb_phs_tab(n_bands, phs_0_pt)
    else:
        phs = np.zeros(n_bands)

    # build multiband modulation function
    n = np.size(pulse_in)
    b = np.zeros(n, dtype="complex")
    for ii in range(0, n_bands):
        b += np.exp(
            1j
            * 2
            * np.pi
            / n
            * band_sep
            * np.arange(-n / 2, n / 2, 1)
            * (ii - (n_bands - 1) / 2)
        ) * np.exp(1j * phs[ii])

    pulse_out = b * pulse_in

    return pulse_out


def mb_phs_tab(n_bands, phs_type="phs_mod"):
    # Return phases to minimize peak b1 amplitude of an MB pulse

    if phs_type == "phs_mod":
        if n_bands < 3 or n_bands > 16:
            raise Exception("Wongs phases valid for 2 < nBands < 17.")

        # Eric Wong's phases: From E C Wong, ISMRM 2012, p. 2209
        p = np.zeros((14, 16))
        p[0, 1:3] = np.array([0.73, 4.602])
        p[1, 1:4] = np.array([3.875, 5.94, 6.197])
        p[2, 1:5] = np.array([3.778, 5.335, 0.872, 0.471])
        p[3, 1:6] = np.array([2.005, 1.674, 5.012, 5.736, 4.123])
        p[4, 1:7] = np.array([3.002, 5.998, 5.909, 2.624, 2.528, 2.440])
        p[5, 1:8] = np.array([1.036, 3.414, 3.778, 3.215, 1.756, 4.555, 2.467])
        p[6, 1:9] = np.array(
            [1.250, 1.783, 3.558, 0.739, 3.319, 1.296, 0.521, 5.332]
        )
        p[7, 1:10] = np.array(
            [4.418, 2.360, 0.677, 2.253, 3.472, 3.040, 3.974, 1.192, 2.510]
        )
        p[8, 1:11] = np.array(
            [
                5.041,
                4.285,
                3.001,
                5.765,
                4.295,
                0.056,
                4.213,
                6.040,
                1.078,
                2.759,
            ]
        )
        p[9, 1:12] = np.array(
            [
                2.755,
                5.491,
                4.447,
                0.231,
                2.499,
                3.539,
                2.931,
                2.759,
                5.376,
                4.554,
                3.479,
            ]
        )
        p[10, 1:13] = np.array(
            [
                0.603,
                0.009,
                4.179,
                4.361,
                4.837,
                0.816,
                5.995,
                4.150,
                0.417,
                1.520,
                4.517,
                1.729,
            ]
        )
        p[11, 1:14] = np.array(
            [
                3.997,
                0.830,
                5.712,
                3.838,
                0.084,
                1.685,
                5.328,
                0.237,
                0.506,
                1.356,
                4.025,
                4.483,
                4.084,
            ]
        )
        p[12, 1:15] = np.array(
            [
                4.126,
                2.266,
                0.957,
                4.603,
                0.815,
                3.475,
                0.977,
                1.449,
                1.192,
                0.148,
                0.939,
                2.531,
                3.612,
                4.801,
            ]
        )
        p[13, 1:16] = np.array(
            [
                4.359,
                3.510,
                4.410,
                1.750,
                3.357,
                2.061,
                5.948,
                3.000,
                2.822,
                0.627,
                2.768,
                3.875,
                4.173,
                4.224,
                5.941,
            ]
        )

        out = p[n_bands - 3, 0:n_bands]

    elif phs_type == "amp_mod":
        # Malik's Hermitian phases: From S J Malik, ISMRM 2015, p. 2398
        if n_bands < 4 or n_bands > 12:
            raise Exception("Maliks phases valid for 3 < nBands < 13.")

        p = np.zeros((9, 12))
        p[0, 0:4] = np.array([0, np.pi, np.pi, 0])
        p[1, 0:5] = np.array([0, 0, np.pi, 0, 0])
        p[2, 0:6] = np.array([1.691, 2.812, 1.157, -1.157, -2.812, -1.691])
        p[3, 0:7] = np.array([2.582, -0.562, 0.102, 0, -0.102, 0.562, -2.582])
        p[4, 0:8] = np.array(
            [2.112, 0.220, 1.464, 1.992, -1.992, -1.464, -0.220, -2.112]
        )
        p[5, 0:9] = np.array(
            [0.479, -2.667, -0.646, -0.419, 0, 0.419, 0.646, 2.667, -0.479]
        )
        p[6, 0:10] = np.array(
            [
                1.683,
                -2.395,
                2.913,
                0.304,
                0.737,
                -0.737,
                -0.304,
                -2.913,
                2.395,
                -1.683,
            ]
        )
        p[7, 0:11] = np.array(
            [
                1.405,
                0.887,
                -1.854,
                0.070,
                -1.494,
                0,
                1.494,
                -0.070,
                1.854,
                -0.887,
                -1.405,
            ]
        )
        p[8, 0:12] = np.array(
            [
                1.729,
                0.444,
                0.722,
                2.190,
                -2.196,
                0.984,
                -0.984,
                2.196,
                -2.190,
                -0.722,
                -0.444,
                -1.729,
            ]
        )

        out = p[n_bands - 4, 0:n_bands]

    elif phs_type == "quad_mod":
        # Grissom's quadratic phases (unpublished)
        k = 3.4 / n_bands  # quadratic phase coefficient
        out = k * (np.arange(0, n_bands, 1) - (n_bands - 1) / 2) ** 2

    else:
        raise Exception('phase type ("{}") not recognized.'.format(phs_type))

    return out


def dz_pins(
    tb,
    sl_sep,
    sl_thick,
    g_max,
    g_slew,
    dt,
    b1_max=0.18,
    ptype="ex",
    ftype="ls",
    d1=0.01,
    d2=0.01,
    gambar=4258,
):
    r"""PINS multiband pulse design.

    Args:
        tb (float): time-bandwidth product.
        sl_sep (float): slice separation in cm.
        sl_thick (float): slice thickness in cm.
        g_max (float): max gradient amplitude in gauss/cm
        g_slew (float): max gradient sliew in gauss/cm/s
        dt (float): RF + gradient dwell time in s.
        b1_max (float): Maximum RF amplitude
        ptype (string): pulse type, 'st' (small-tip excitation), 'ex' (pi/2
            excitation pulse), 'se' (spin-echo pulse), 'inv' (inversion), or
            'sat' (pi/2 saturation pulse).
        ftype (string): type of filter to use in pulse design
        d1 (float): passband ripple level in :math:'M_0^{-1}'.
        d2 (float): stopband ripple level in :math:'M_0^{-1}'.
        gambar (float): Appropriate gyromagnetic ratio in Hz/gauss.

    Returns:
        2-element tuple containing:

        - **rf** (*array*): RF Pulse in Gauss
        - **g** (*array*): Gradient waveform in Gauss/cm

    References:
        Norris, D.G. and Koopmans, P.J. and Boyacioglu, R and Barth, M (2011).
        'Power independent of number of slices (PINS) radiofrequency Pulses
        for low-power simultaneous multislice excitation'.
        Magn. Reson. Med., 66(5):1234-1240.
    """

    kz_width = tb / sl_thick  # 1/cm, width in k-space we must go
    # calcualte number of subpulses (odd)
    n_pulses = int(2 * np.floor(np.ceil(kz_width / (1 / sl_sep)) / 2))
    # call SLR to get envelope
    rf_soft = slr.dzrf(n_pulses, tb, ptype, ftype, d1, d2)

    # design the blip trapezoid
    g_area = 1 / sl_sep / gambar
    [gz_blip, _] = trap_grad(g_area, g_max, g_slew, dt)

    # Calculate the block/hard RF pulse width based on
    b1_scaled = 2 * np.pi * gambar * b1_max * dt
    hpw = int(np.ceil(np.max(np.abs(rf_soft)) / b1_scaled))

    # interleave RF subpusles with gradient subpulses to form full pulses
    rf = np.kron(
        rf_soft[:-1],
        np.concatenate((np.ones(hpw), np.zeros((np.size(gz_blip))))),
    )
    rf = np.concatenate((rf, rf_soft[-1] * np.ones(hpw)))
    rf = rf / (np.sum(rf) * 2 * np.pi * gambar * dt) * np.sum(rf_soft)

    g = np.concatenate([np.zeros(hpw), np.squeeze(gz_blip)])
    g = np.tile(g, n_pulses - 1)
    g = np.concatenate((g, np.zeros(hpw)))

    return rf, g



"""This MRI submodule contains functions and classes for MRI pulse design.

It contains functions to design a variety of RF pulses for MRI, such as SLR,
adiabatic, parallel transmit, multibanded, and others. The submodule also
includes other functions to assist with pulse design, such as I/O functions,
trajectory/gradient designers, and Bloch simulators.

Explore RF design tutorials at `sigpy-rf-tutorials`_. These are primarily
Jupyter Notebooks, and provide more detailed instruction on pulse design
workflow and function use.

See in-progress features at `sigpy-rf`_.

.. _sigpy-rf-tutorials: https://github.com/jonbmartin/sigpy-rf-tutorials
.. _sigpy-rf: https://github.com/jonbmartin/sigpy-rf

"""
from sigpy.mri import linop
from sigpy.mri.rf import (
    adiabatic,
    b1sel,
    io,
    multiband,
    optcont,
    ptx,
    shim,
    sim,
    slr,
    trajgrad,
    util,
)
from sigpy.mri.rf.adiabatic import *  # noqa
from sigpy.mri.rf.b1sel import *  # noqa
from sigpy.mri.rf.io import *  # noqa
from sigpy.mri.rf.linop import *  # noqa
from sigpy.mri.rf.multiband import *  # noqa
from sigpy.mri.rf.optcont import *  # noqa
from sigpy.mri.rf.ptx import *  # noqa
from sigpy.mri.rf.shim import *  # noqa
from sigpy.mri.rf.sim import *  # noqa
from sigpy.mri.rf.slr import *  # noqa
from sigpy.mri.rf.trajgrad import *  # noqa
from sigpy.mri.rf.util import *  # noqa

__all__ = ["linop"]
__all__.extend(adiabatic.__all__)
__all__.extend(b1sel.__all__)
__all__.extend(io.__all__)
__all__.extend(multiband.__all__)
__all__.extend(optcont.__all__)
__all__.extend(ptx.__all__)
__all__.extend(sim.__all__)
__all__.extend(shim.__all__)
__all__.extend(slr.__all__)
__all__.extend(trajgrad.__all__)
__all__.extend(util.__all__)



import unittest

import numpy as np
import numpy.testing as npt

from sigpy import app, linop, prox, util

if __name__ == "__main__":
    unittest.main()


class TestApp(unittest.TestCase):
    def test_MaxEig(self):
        n = 5
        mat = util.randn([n, n])
        A = linop.MatMul([n, 1], mat)
        s = np.linalg.svd(mat, compute_uv=False)

        npt.assert_allclose(
            app.MaxEig(A.H * A, max_iter=1000, show_pbar=False).run(),
            s[0] ** 2,
            atol=1e-3,
        )

    def test_LinearLeastSquares(self):
        n = 5
        _A = np.eye(n) + 0.1 * np.ones([n, n])
        A = linop.MatMul([n, 1], _A)
        x = np.arange(n).reshape([n, 1])
        y = A(x)

        for z in [None, x.copy()]:
            for lamda in [0, 0.1]:
                for proxg in [None, prox.L2Reg([n, 1], lamda)]:
                    for solver in [
                        "GradientMethod",
                        "PrimalDualHybridGradient",
                        "ConjugateGradient",
                        "ADMM",
                    ]:
                        with self.subTest(
                            proxg=proxg, solver=solver, lamda=lamda, z=z
                        ):
                            AHA = _A.T @ _A + lamda * np.eye(n)
                            AHy = _A.T @ y
                            if proxg is not None:
                                AHA += lamda * np.eye(n)

                            if z is not None:
                                AHy = _A.T @ y + lamda * z

                            x_numpy = np.linalg.solve(AHA, AHy)
                            if (
                                solver == "ConjugateGradient"
                                and proxg is not None
                            ):
                                with self.assertRaises(ValueError):
                                    app.LinearLeastSquares(
                                        A,
                                        y,
                                        solver=solver,
                                        lamda=lamda,
                                        proxg=proxg,
                                        z=z,
                                        show_pbar=False,
                                    ).run()
                            else:
                                x_rec = app.LinearLeastSquares(
                                    A,
                                    y,
                                    solver=solver,
                                    lamda=lamda,
                                    proxg=proxg,
                                    z=z,
                                    show_pbar=False,
                                ).run()

                                npt.assert_allclose(x_rec, x_numpy, atol=1e-3)

    def test_precond_LinearLeastSquares(self):
        n = 5
        _A = np.eye(n) + 0.01 * util.randn([n, n])
        A = linop.MatMul([n, 1], _A)
        x = util.randn([n, 1])
        y = A(x)
        x_lstsq = np.linalg.lstsq(_A, y, rcond=-1)[0]
        p = 1 / (np.sum(abs(_A) ** 2, axis=0).reshape([n, 1]))

        P = linop.Multiply([n, 1], p)
        x_rec = app.LinearLeastSquares(A, y, show_pbar=False).run()
        npt.assert_allclose(x_rec, x_lstsq, atol=1e-3)

        alpha = 1 / app.MaxEig(P * A.H * A, show_pbar=False).run()
        x_rec = app.LinearLeastSquares(
            A,
            y,
            solver="GradientMethod",
            alpha=alpha,
            max_power_iter=100,
            max_iter=1000,
            show_pbar=False,
        ).run()
        npt.assert_allclose(x_rec, x_lstsq, atol=1e-3)

        tau = p
        x_rec = app.LinearLeastSquares(
            A,
            y,
            solver="PrimalDualHybridGradient",
            max_iter=1000,
            tau=tau,
            show_pbar=False,
        ).run()
        npt.assert_allclose(x_rec, x_lstsq, atol=1e-3)

    def test_dual_precond_LinearLeastSquares(self):
        n = 5
        _A = np.eye(n) + 0.1 * util.randn([n, n])
        A = linop.MatMul([n, 1], _A)
        x = util.randn([n, 1])
        y = A(x)
        x_lstsq = np.linalg.lstsq(_A, y, rcond=-1)[0]

        d = 1 / np.sum(abs(_A) ** 2, axis=1, keepdims=True).reshape([n, 1])
        x_rec = app.LinearLeastSquares(
            A,
            y,
            solver="PrimalDualHybridGradient",
            max_iter=1000,
            sigma=d,
            show_pbar=False,
        ).run()
        npt.assert_allclose(x_rec, x_lstsq, atol=1e-3)



# -*- coding: utf-8 -*-
"""Optimal Control Pulse Design functions.
"""
from sigpy import backend

__all__ = ["blochsim", "deriv"]


def blochsim(rf, x, g):
    r"""1D RF pulse simulation, with simultaneous RF + gradient rotations.
    Assume x has inverse spatial units of g, and g has gamma*dt applied and
    assume x = [...,Ndim], g = [Ndim,Nt].

     Args:
         rf (array): rf waveform input.
         x (array): spatial locations.
         g (array): gradient waveform.

     Returns:
         array: SLR alpha parameter
         array: SLR beta parameter
    """

    device = backend.get_device(rf)
    xp = device.xp
    with device:
        a = xp.ones(xp.shape(x)[0], dtype=complex)
        b = xp.zeros(xp.shape(x)[0], dtype=complex)
        for mm in range(0, xp.size(rf), 1):  # loop over time
            # apply RF
            c = xp.cos(xp.abs(rf[mm]) / 2)
            s = 1j * xp.exp(1j * xp.angle(rf[mm])) * xp.sin(xp.abs(rf[mm]) / 2)
            at = a * c - b * xp.conj(s)
            bt = a * s + b * c
            a = at
            b = bt

            # apply gradient
            if g.ndim > 1:
                z = xp.exp(-1j * x @ g[mm, :])
            else:
                z = xp.exp(-1j * x * g[mm])
            b = b * z

        # apply total phase accrual
        if g.ndim > 1:
            z = xp.exp(1j / 2 * x @ xp.sum(g, 0))
        else:
            z = xp.exp(1j / 2 * x * xp.sum(g))
        a = a * z
        b = b * z

        return a, b


def deriv(rf, x, g, auxa, auxb, af, bf):
    r"""1D RF pulse simulation, with simultaneous RF + gradient rotations.

    'rf', 'g', and 'x' should have consistent units.

     Args:
         rf (array): rf waveform input.
         x (array): spatial locations.
         g (array): gradient waveform.
         auxa (None or array): auxa
         auxb (array): auxb
         af (array): forward sim a.
         bf( array): forward sim b.

     Returns:
         array: SLR alpha parameter
         array: SLR beta parameter
    """

    device = backend.get_device(rf)
    xp = device.xp
    with device:
        drf = xp.zeros(xp.shape(rf), dtype=complex)
        ar = xp.ones(xp.shape(af), dtype=complex)
        br = xp.zeros(xp.shape(bf), dtype=complex)

        for mm in range(xp.size(rf) - 1, -1, -1):
            # calculate gradient blip phase
            if g.ndim > 1:
                z = xp.exp(1j / 2 * x @ g[mm, :])
            else:
                z = xp.exp(1j / 2 * x * g[mm])

            # strip off gradient blip from forward sim
            af = af * xp.conj(z)
            bf = bf * z

            # add gradient blip to backward sim
            ar = ar * z
            br = br * z

            # strip off the curent rf rotation from forward sim
            c = xp.cos(xp.abs(rf[mm]) / 2)
            s = 1j * xp.exp(1j * xp.angle(rf[mm])) * xp.sin(xp.abs(rf[mm]) / 2)
            at = af * c + bf * xp.conj(s)
            bt = -af * s + bf * c
            af = at
            bf = bt

            # calculate derivatives wrt rf[mm]
            db1 = xp.conj(1j / 2 * br * bf) * auxb
            db2 = xp.conj(1j / 2 * af) * ar * auxb
            drf[mm] = xp.sum(db2 + xp.conj(db1))
            if auxa is not None:
                da1 = xp.conj(1j / 2 * bf * ar) * auxa
                da2 = 1j / 2 * xp.conj(af) * br * auxa
                drf[mm] += xp.sum(da2 + xp.conj(da1))

            # add current rf rotation to backward sim
            art = ar * c - xp.conj(br) * s
            brt = br * c + xp.conj(ar) * s
            ar = art
            br = brt

        return drf



import unittest

import numpy as np
import numpy.testing as npt

import sigpy.mri.rf as rf

if __name__ == "__main__":
    unittest.main()


class TestMultiband(unittest.TestCase):
    def test_multiband(self):
        # slr pulse
        tb = 8
        N = 512
        d1 = 0.01
        d2 = 0.01
        p_type = "ex"
        f_type = "ls"
        pulse = rf.slr.dzrf(N, tb, p_type, f_type, d1, d2, True)

        # multiband it
        n_bands = 3
        phs_type = "phs_mod"  # phsMod, ampMod, or quadMod
        band_sep = 5 * tb  # separate by 5 slice widths
        mb_pulse = rf.multiband.mb_rf(pulse, n_bands, band_sep, phs_type)

        # simulate it
        [a, b] = rf.sim.abrm(
            mb_pulse, np.arange(-20 * tb, 20 * tb, 40 * tb / 2000), True
        )
        mxy = 2 * np.multiply(np.conj(a), b)

        pts = np.array([mxy[750], mxy[850], mxy[1000], mxy[1150], mxy[1250]])
        npt.assert_almost_equal(abs(pts), np.array([1, 0, 1, 0, 1]), decimal=2)

    def test_pins(self):
        # pins pulse specs
        tb = 8
        d1 = 0.01
        d2 = 0.01
        p_type = "ex"
        f_type = "ls"

        sl_sep = 3  # cm
        sl_thick = 0.3  # cm
        g_max = 4  # gauss/cm
        g_slew = 18000  # gauss/cm/s
        dt = 4e-6  # seconds, dwell time
        b1_max = 0.18  # gauss
        [rf_pins, g_pins] = rf.multiband.dz_pins(
            tb,
            sl_sep,
            sl_thick,
            g_max,
            g_slew,
            dt,
            b1_max,
            p_type,
            f_type,
            d1,
            d2,
        )

        # simulate it
        x = np.reshape(np.arange(-1000, 1000), (2000, 1)) / 1000 * 12  # cm
        [a, b] = rf.sim.abrm_nd(
            2 * np.pi * dt * 4258 * rf_pins,
            x,
            np.reshape(g_pins, (np.size(g_pins), 1)) * 4258 * dt * 2 * np.pi,
        )
        mxy = 2 * np.conj(a) * b

        pts = np.array([mxy[100], mxy[1000], mxy[1900]])
        npt.assert_almost_equal(abs(pts), np.array([0, 1, 0]), decimal=2)
